#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import os
import os.path
import select
import socket
import time

from fenrirscreenreader.core import debug
from fenrirscreenreader.core import remoteInstanceRegistry
from fenrirscreenreader.core.eventData import FenrirEventType
from fenrirscreenreader.core.remoteDriver import RemoteDriver as remoteDriver


MAIN_SOCKET_FILE = "/tmp/fenrirscreenreader-deamon.sock"


class driver(remoteDriver):
    def __init__(self):
        remoteDriver.__init__(self)
        self.fenrirSocks = []
        self.bound_sockets = []

    def initialize(self, environment):
        self.env = environment
        # Use threading instead of multiprocessing to avoid pickle issues
        # with self.env (which contains unpicklable file handles)
        self.env["runtime"]["ProcessManager"].add_custom_event_thread(
            self.watch_dog, multiprocess=False
        )

    def _get_configured_socket_file(self):
        socket_file = ""
        try:
            socket_file = self.env["runtime"]["SettingsManager"].get_setting(
                "remote", "socket_file"
            )
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "unixDriver watch_dog: Error getting socket file setting: "
                + str(e),
                debug.DebugLevel.ERROR,
            )
        return socket_file

    def _get_socket_candidates(self):
        configured_socket_file = self._get_configured_socket_file()
        if configured_socket_file:
            return [(configured_socket_file, False)]

        screen_driver = self.env["runtime"]["SettingsManager"].get_setting(
            "screen", "driver"
        )
        if screen_driver == "vcsaDriver":
            return [(MAIN_SOCKET_FILE, False)]

        private_socket_file = (
            "/tmp/fenrirscreenreader-" + str(os.getpid()) + ".sock"
        )
        return [(private_socket_file, False), (MAIN_SOCKET_FILE, True)]

    def _is_socket_active(self, socket_file):
        test_sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            test_sock.settimeout(0.2)
            test_sock.connect(socket_file)
            return True
        except OSError:
            return False
        finally:
            test_sock.close()

    def _bind_socket(self, socket_file, optional):
        if os.path.exists(socket_file):
            if optional and self._is_socket_active(socket_file):
                return None
            os.unlink(socket_file)

        fenrir_sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            fenrir_sock.bind(socket_file)
            os.chmod(socket_file, 0o666)  # Allow all users to read/write
            fenrir_sock.listen(1)
        except OSError:
            fenrir_sock.close()
            if optional:
                return None
            raise
        return fenrir_sock

    def _register_instance(self):
        settings_manager = self.env["runtime"]["SettingsManager"]
        active_socket_files = [
            socket_file
            for fenrir_sock, socket_file in self.bound_sockets
            if self._socket_file_matches_socket(fenrir_sock, socket_file)
        ]
        instance_data = {
            "pid": os.getpid(),
            "ppid": os.getppid(),
            "socket_files": active_socket_files,
            "main_socket": MAIN_SOCKET_FILE in active_socket_files,
            "screen_driver": settings_manager.get_setting("screen", "driver"),
            "keyboard_driver": settings_manager.get_setting("keyboard", "driver"),
            "x11_window_id": settings_manager.get_setting(
                "keyboard", "x11_window_id"
            ),
            "created_at": time.time(),
        }
        remoteInstanceRegistry.write_instance(instance_data)

    def _socket_file_matches_socket(self, fenrir_sock, socket_file):
        try:
            if os.stat(socket_file).st_uid != os.getuid():
                return False
            fd_stat = os.fstat(fenrir_sock.fileno())
            socket_inode = str(fd_stat.st_ino)
            with open("/proc/net/unix", "r", encoding="utf-8") as proc_file:
                for line in proc_file:
                    line_parts = line.split()
                    if (
                        len(line_parts) >= 8
                        and line_parts[6] == socket_inode
                        and line_parts[7] == socket_file
                    ):
                        return True
        except OSError:
            return False
        return False

    def _cleanup(self):
        for fenrir_sock in self.fenrirSocks:
            try:
                fenrir_sock.close()
            except OSError:
                pass
        self.fenrirSocks = []

        for _, socket_file in self.bound_sockets:
            try:
                if (
                    os.path.exists(socket_file)
                    and not self._is_socket_active(socket_file)
                ):
                    os.unlink(socket_file)
            except OSError:
                pass
        self.bound_sockets = []
        remoteInstanceRegistry.remove_instance()

    def _handle_client(self, client_sock, event_queue):
        try:
            rawdata = client_sock.recv(8129)
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "unixDriver watch_dog: Error receiving data from client: "
                + str(e),
                debug.DebugLevel.ERROR,
            )
            rawdata = b""

        try:
            data = rawdata.decode("utf-8").rstrip().lstrip()
            upper_data = data.upper()
            if upper_data in ("LS", "LIST", "COMMAND LS", "COMMAND LIST"):
                response = self.env["runtime"][
                    "RemoteManager"
                ].handle_remote_incomming_with_response(data)
                client_sock.sendall((response["message"] + "\n").encode("utf-8"))
                return

            event_queue.put(
                {
                    "Type": FenrirEventType.remote_incomming,
                    "data": data,
                }
            )
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "unixDriver watch_dog: Error decoding/queuing data: "
                + str(e),
                debug.DebugLevel.ERROR,
            )

    def watch_dog(self, active, event_queue):
        # echo "command say this is a test" | socat -
        # UNIX-CLIENT:/tmp/fenrirscreenreader-deamon.sock
        for socket_file, optional in self._get_socket_candidates():
            fenrir_sock = self._bind_socket(socket_file, optional)
            if fenrir_sock is None:
                continue
            self.fenrirSocks.append(fenrir_sock)
            self.bound_sockets.append((fenrir_sock, socket_file))

        if not self.fenrirSocks:
            return

        self._register_instance()
        last_register = time.time()
        while active.value:
            if time.time() - last_register > 10.0:
                self._register_instance()
                last_register = time.time()

            # Check if the client is still connected and if data is available:
            try:
                r, _, _ = select.select(self.fenrirSocks, [], [], 0.8)
            except select.error:
                break
            if r == []:
                continue
            for fenrir_sock in r:
                client_sock, client_addr = fenrir_sock.accept()
                # Ensure client socket is always closed to prevent resource
                # leaks
                try:
                    self._handle_client(client_sock, event_queue)
                finally:
                    # Always close client socket, even if data processing fails
                    try:
                        client_sock.close()
                    except Exception as e:
                        self.env["runtime"]["DebugManager"].write_debug_out(
                            "unixDriver watch_dog: Error closing client socket: "
                            + str(e),
                            debug.DebugLevel.ERROR,
                        )
        self._cleanup()

    def shutdown(self):
        self._cleanup()
        remoteDriver.shutdown(self)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import re
import string

from fenrirscreenreader.core import debug


class BarrierManager:
    def __init__(self):
        self.currIsBarrier = False
        self.prefIsBarrier = False

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def update_barrier_change(self, is_barrier):
        self.prefIsBarrier = self.currIsBarrier
        self.currIsBarrier = is_barrier

    def reset_barrier_change(self):
        self.currIsBarrier = False
        self.prefIsBarrier = False

    def is_barrier_change(self):
        return self.currIsBarrier != self.prefIsBarrier

    def handle_line_barrier(
        self, text, xCursor, yCursor, output=True, do_interrupt=True
    ):
        is_barrier = False
        try:
            is_barrier, say_line = self.get_barrier_text(
                text, xCursor, yCursor
            )
        except Exception as e:
            return False, ""

        self.update_barrier_change(is_barrier)
        if self.is_barrier_change():
            if output:
                if is_barrier:
                    self.env["runtime"]["OutputManager"].play_sound_icon(
                        sound_icon="BarrierStart", interrupt=do_interrupt
                    )
                else:
                    self.env["runtime"]["OutputManager"].play_sound_icon(
                        sound_icon="BarrierEnd", interrupt=do_interrupt
                    )

        if not is_barrier:
            say_line = ""
        return is_barrier, say_line

    def get_barrier_text(self, text, xCursor, yCursor):
        line = text[yCursor]
        if not self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "barrier", "enabled"
        ):
            return False, line
        offset = xCursor

        left_barriers = self.env["runtime"]["SettingsManager"].get_setting(
            "barrier", "left_barriers"
        )
        right_barriers = self.env["runtime"]["SettingsManager"].get_setting(
            "barrier", "right_barriers"
        )
        # is the cursor at the begin or end of an entry:
        # print(line[:offset + 1].count('│'),line[offset:].count('│'))
        # start

        for b in left_barriers:
            if line[: offset + 1].count(b) > line[offset:].count(b):
                offset = xCursor - 1
            start = line[:offset].rfind(b)
            if start != -1:
                start += 1
                break
        if start == -1:
            return False, line
        # end
        for b in right_barriers:
            end = line[start:].find(b)
            if end != -1:
                end = start + end
                break
        if end == -1:
            return False, line
        if start == end:
            return False, line

        return True, line[start:end]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.core.tabCompletionManager import TabCompletionManager


class command:
    def __init__(self):
        self.manager = TabCompletionManager()

    def initialize(self, environment):
        self.env = environment
        self.manager.initialize(environment)

    def shutdown(self):
        pass

    def get_description(self):
        return _("Tracks tab keypresses for completion announcements")

    def run(self):
        self.manager.capture_if_tab()

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3

import importlib.util
import os

from fenrirscreenreader.core.i18n import _

# Load base configuration class
_base_path = os.path.join(os.path.dirname(__file__), "..", "config_base.py")
_spec = importlib.util.spec_from_file_location("config_base", _base_path)
_module = importlib.util.module_from_spec(_spec)
_spec.loader.exec_module(_module)
config_command = _module.config_command


class command(config_command):
    def __init__(self):
        super().__init__()

    def get_description(self):
        return "Toggle debug mode"

    def run(self):
        current_level = self.get_setting("general", "debug_level", "0")

        # Present current debug level
        if current_level == "0":
            self.present_text("Current debug mode: disabled")
            new_level = "1"
            state_text = "enabled"
        else:
            self.present_text(f"Current debug level: {current_level}")
            new_level = "0"
            state_text = "disabled"

        success = self.set_setting("general", "debug_level", new_level)

        if success:
            self.present_text(f"Debug mode {state_text}")
            if new_level != "0":
                debug_mode = self.get_setting("general", "debug_mode", "File")
                if debug_mode == "File":
                    self.present_text(
                        "Debug output will be written to log file"
                    )
                else:
                    self.present_text(
                        "Debug output will be printed to console"
                    )
            self.play_sound("Accept")
        else:
            self.present_text("Failed to change debug mode")
            self.play_sound("Error")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from enum import Enum

from fenrirscreenreader.core import debug


class FenrirEventType(Enum):
    ignore = 0
    stop_main_loop = 1
    screen_update = 2
    keyboard_input = 3
    plug_input_device = 4
    screen_changed = 5
    heart_beat = 6
    execute_command = 7
    remote_incomming = 8

    def __int__(self):
        return self.value

    def __str__(self):
        return self.name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.utils import line_utils


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _(
            "read line to cursor pos, use review cursor if you are in review mode, "
            "otherwhise use text cursor"
        )

    def run(self):
        cursor_pos = self.env["runtime"][
            "CursorManager"
        ].get_review_or_text_cursor()

        x, y, curr_line = line_utils.get_current_line(
            cursor_pos["x"],
            cursor_pos["y"],
            self.env["screen"]["new_content_text"],
        )

        if curr_line.isspace():
            self.env["runtime"]["OutputManager"].present_text(
                _("blank"), sound_icon="EmptyLine", interrupt=True
            )
        else:
            self.env["runtime"]["OutputManager"].present_text(
                curr_line[: cursor_pos["x"]], interrupt=True
            )
        self.env["runtime"]["OutputManager"].announce_active_cursor()

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.


from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return ""

    def run(self):
        if self.env["runtime"]["PunctuationManager"].cycle_punctuation():
            self.env["runtime"]["OutputManager"].present_text(
                self.env["runtime"]["SettingsManager"].get_setting(
                    "general", "punctuation_level"
                ),
                interrupt=True,
                ignore_punctuation=True,
            )
        else:
            self.env["runtime"]["OutputManager"].present_text(
                _("No punctuation found."),
                interrupt=True,
                ignore_punctuation=True,
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core import debug


def get_prev_char(curr_x, curr_y, curr_text):
    line_break = False
    end_of_screen = False
    if curr_text == "":
        return -1, -1, "", end_of_screen, line_break
    wrapped_lines = curr_text.split("\n")
    x = curr_x
    y = curr_y
    if x - 1 < 0:
        if y - 1 >= 0:
            y -= 1
            x = len(wrapped_lines[y]) - 1
            line_break = True
        else:
            line_break = False
            end_of_screen = True
    else:
        x -= 1
    curr_char = ""
    if not end_of_screen:
        curr_char = wrapped_lines[y][x]
    return x, y, curr_char, end_of_screen, line_break


def get_current_char(curr_x, curr_y, curr_text):
    if curr_text == "":
        return -1, -1, ""
    wrapped_lines = curr_text.split("\n")
    curr_char = wrapped_lines[curr_y][curr_x]
    return curr_x, curr_y, curr_char


def get_up_char(curr_x, curr_y, curr_text):
    end_of_screen = False
    if curr_text == "":
        return -1, -1, "", end_of_screen
    wrapped_lines = curr_text.split("\n")
    curr_y -= 1
    if curr_y < 0:
        curr_y = 0
    else:
        end_of_screen = True
    curr_char = ""
    if not end_of_screen:
        curr_char = wrapped_lines[curr_y][curr_x]
    return curr_x, curr_y, curr_char, end_of_screen


def get_down_char(curr_x, curr_y, curr_text):
    end_of_screen = False
    if curr_text == "":
        return -1, -1, "", end_of_screen
    wrapped_lines = curr_text.split("\n")
    curr_y += 1
    if curr_y >= len(wrapped_lines):
        curr_y = len(wrapped_lines) - 1
    else:
        end_of_screen = True
    curr_char = ""
    if not end_of_screen:
        curr_char = wrapped_lines[curr_y][curr_x]
    return curr_x, curr_y, curr_char, end_of_screen


def get_last_char_in_line(curr_y, curr_text):
    end_of_screen = False
    if curr_text == "":
        return -1, -1, ""
    wrapped_lines = curr_text.split("\n")
    curr_x = len(wrapped_lines[curr_y].rstrip()) - 1
    if curr_x < 0:
        curr_x = 0
    curr_char = wrapped_lines[curr_y][curr_x]
    return curr_x, curr_y, curr_char


def get_next_char(curr_x, curr_y, curr_text):
    line_break = False
    end_of_screen = False
    if curr_text == "":
        return -1, -1, "", end_of_screen, line_break
    wrapped_lines = curr_text.split("\n")
    x = curr_x
    y = curr_y
    if x + 1 == len(wrapped_lines[y]):
        if y + 1 <= len(wrapped_lines) - 1:
            y += 1
            x = 0
            line_break = True
        else:
            line_break = False
            end_of_screen = True
    else:
        x += 1
    curr_char = ""
    if not end_of_screen:
        curr_char = wrapped_lines[y][x]

    return x, y, curr_char, end_of_screen, line_break


def get_phonetic(curr_char):
    if len(curr_char) != 1:
        return curr_char
    phonetics_dict = {
        "A": "alpha",
        "B": "bravo",
        "C": "charlie",
        "D": "delta",
        "E": "echo",
        "F": "foxtrot",
        "G": "golf",
        "H": "hotel",
        "I": "india",
        "J": "juliet",
        "K": "kilo",
        "L": "lima",
        "M": "mike",
        "N": "november",
        "O": "oscar",
        "P": "papa",
        "Q": "quebec",
        "R": "romeo",
        "S": "sierra",
        "T": "tango",
        "U": "uniform",
        "V": "victor",
        "W": "whisky",
        "X": "x ray",
        "Y": "yankee",
        "Z": "zulu",
    }
    try:
        phon_char = phonetics_dict[curr_char.upper()]
        if curr_char.isupper():
            phon_char = phon_char[0].upper() + phon_char[1:]
        return phon_char
    except Exception as e:
        # Utility function, no env access - return fallback
        return curr_char


def present_char_for_review(
    env, char, interrupt=True, announce_capital=True, flush=False
):
    """Present a character for explicit review commands only"""
    if char == " ":
        if " " in env["punctuation"]["PUNCTDICT"]:
            announce_char = env["punctuation"]["PUNCTDICT"][" "]
        else:
            announce_char = "space"
        env["runtime"]["OutputManager"].present_text(
            announce_char, interrupt=interrupt, flush=flush
        )
    else:
        env["runtime"]["OutputManager"].present_text(
            char,
            interrupt=interrupt,
            ignore_punctuation=True,
            announce_capital=announce_capital,
            flush=flush,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core import debug

settings_data = {
    "sound": {
        "enabled": True,
        "driver": "genericDriver",
        "theme": "default",
        "volume": 1.0,
        "generic_play_file_command": "play -q -v fenrirVolume fenrirSoundFile",
        "generic_frequency_command": "play -q -v fenrirVolume -n -c1 synth fenrirDuration sine fenrirFrequence",
        "progress_monitoring": True,
    },
    "speech": {
        "enabled": True,
        "driver": "genericDriver",
        "server_path": "",
        "rate": 0.75,
        "pitch": 0.5,
        "capital_pitch": 0.8,
        "capital_indicator": "pitch",
        "volume": 1.0,
        "module": "",
        "voice": "en-us",
        "language": "",
        "auto_read_incoming": True,
        "read_numbers_as_digits": False,
        "rapid_update_threshold": 5,
        "rapid_update_window": 0.3,
        "batch_flush_interval": 0.5,
        "max_batch_lines": 100,
        "flood_line_threshold": 500,
        "generic_speech_command": 'espeak -a fenrirVolume -s fenrirRate -p fenrirPitch -v fenrirVoice "fenrirText"',
        "fenrir_min_volume": 0,
        "fenrir_max_volume": 200,
        "fenrir_min_pitch": 0,
        "fenrir_max_pitch": 99,
        "fenrir_min_rate": 80,
        "fenrir_max_rate": 450,
    },
    "screen": {
        "driver": "vcsaDriver",
        "encoding": "auto",
        "screen_update_delay": 0.1,
        "ignore_screen": "",
        "autodetect_ignore_screen": False,
    },
    "general": {
        "debug_level": debug.DebugLevel.DEACTIVE,
        "debug_mode": "FILE",
        "debug_file": "",
        "punctuation_profile": "default",
        "punctuation_level": "some",
        "respect_punctuation_pause": True,
        "replace_undefined_punctuation_with_space": True,
        "new_line_pause": True,
        "number_of_clipboards": 10,
        "clipboard_export_path": "/tmp/fenrirClipboard",
        "emoticons": True,
        "fenrir_keys": "KEY_KP0,KEY_META",
        "script_keys": "KEY_COMPOSE",
        "time_format": "%I:%M%P",
        "date_format": "%A, %B %d, %Y",
        "auto_spell_check": False,
        "spell_check_language": "en_US",
        "script_path": "/usr/share/fenrirscreenreader/scripts",
        "command_path": "/usr/share/fenrirscreenreader/commands",
        "attribute_format_string": "Background fenrirBGColor,Foreground fenrirFGColor,fenrirUnderline,fenrirBold,fenrirBlink, Font fenrirFont,Fontsize fenrirFontSize",
        "auto_present_indent": False,
        "auto_present_indent_mode": 1,
        "has_attributes": True,
        "shell": "",
        "diff_presentation": "both",
        "diff_verbosity": "compact",
    },
    "focus": {
        "cursor": True,
        "highlight": False,
        "tui": False,
    },
    "remote": {
        "enabled": True,
        "driver": "unixDriver",
        "port": 22447,
        "socket_file": "",
        "enable_settings_remote": True,
        "enable_command_remote": True,
    },
    "barrier": {
        "enabled": True,
        "left_barriers": "│└┌─",
        "right_barriers": "│┘┐─",
    },
    "review": {
        "line_break": True,
        "end_of_screen": True,
        "leave_review_on_cursor_change": True,
        "leave_review_on_screen_change": True,
    },
    "menu": {
        "vmenu_path": "",
        "quick_menu": "speech#rate;speech#pitch;speech#volume",
    },
    "time": {
        "enabled": False,
        "present_time": True,
        "present_date": True,
        "delay_sec": 0,
        "on_minutes": "00,30",
        "announce": True,
        "interrupt": False,
    },
    "keyboard": {
        "driver": "evdev",
        "device": "all",
        "grab_devices": True,
        "x11_window_id": "",
        "ignore_shortcuts": False,
        "keyboard_layout": "desktop",
        "char_echo_mode": 2,  # while capslock
        "char_delete_echo": True,
        "word_echo": True,
        "interrupt_on_key_press": True,
        "interrupt_on_key_press_filter": "",
        "double_tap_timeout": 0.2,
    },
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import os
import string

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.utils import word_utils

initialized = False
try:
    import enchant

    initialized = True
except Exception as e:
    pass


class command:
    def __init__(self):
        self.language = ""
        self.spellChecker = ""

    def initialize(self, environment):
        self.env = environment
        self.update_spell_language()

    def shutdown(self):
        pass

    def get_description(self):
        return "No Description found"

    def update_spell_language(self):
        if not initialized:
            return
        self.spellChecker = enchant.Dict(
            self.env["runtime"]["SettingsManager"].get_setting(
                "general", "spell_check_language"
            )
        )
        self.language = self.env["runtime"]["SettingsManager"].get_setting(
            "general", "spell_check_language"
        )

    def run(self):
        if not initialized:
            return
        if not self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "general", "auto_spell_check"
        ):
            return
        if (
            self.env["runtime"]["SettingsManager"].get_setting(
                "general", "spell_check_language"
            )
            != self.language
        ):
            try:
                self.update_spell_language()
            except Exception as e:
                return

        # just when horizontal cursor move worddetection is needed
        if not self.env["runtime"][
            "CursorManager"
        ].is_cursor_horizontal_move():
            return

        # for now no new line
        if self.env["runtime"]["CursorManager"].is_cursor_vertical_move():
            return
        # more than a keyecho?
        if len(self.env["screen"]["new_delta"]) > 1:
            return
        # deletion
        if self.env["runtime"]["ScreenManager"].is_negative_delta():
            return
        # first place could not be the end of a word
        if self.env["screen"]["new_cursor"]["x"] == 0:
            return

        # get the word (just for speedup only look at current line
        new_content = self.env["screen"]["new_content_text"].split("\n")[
            self.env["screen"]["new_cursor"]["y"]
        ]
        x, y, curr_word, end_of_screen, line_break = (
            word_utils.get_current_word(
                self.env["screen"]["new_cursor"]["x"], 0, new_content
            )
        )
        # was this a typed word?
        if self.env["runtime"]["ScreenManager"].is_delta():
            if not (
                new_content[self.env["screen"]["old_cursor"]["x"]]
                in string.whitespace + '!"#$%&()*+,-./:;<=>?@[\\]^_{|}~'
                and x != self.env["screen"]["old_cursor"]["x"]
            ):
                return
            else:
                curr_word = curr_word.strip(
                    string.whitespace + '!"#$%&()*+,-./:;<=>?@[\\]^_{|}~'
                )
        else:
            # or just arrow arround?
            if not new_content[
                self.env["screen"]["new_cursor"]["x"]
            ].isspace():
                return
            if (
                x + len(curr_word) != self.env["screen"]["new_cursor"]["x"]
            ) and (
                x + len(curr_word) != self.env["screen"]["new_cursor"]["x"] - 1
            ):
                return

        # just on end of word
        if self.env["screen"]["new_cursor"]["x"] > 0:
            if (
                not new_content[
                    self.env["screen"]["old_cursor"]["x"] - 1
                ].lower()
                in string.ascii_lowercase
            ):
                return

        # ignore bash buildins
        if curr_word in [
            "cd",
            "fg",
            "bg",
            "alias",
            "bind",
            "dir",
            "caller",
            "buildin",
            "command",
            "declare",
            "echo",
            "enable",
            "help",
            "let",
            "local",
            "logout",
            "mapfile",
            "printf",
            "read",
            "readarray",
            "source",
            "type",
            "typeset",
            "ulimit",
            "unalias",
        ]:
            return
        # ignore the application name
        if curr_word.upper() == "FENRIR":
            return
        if curr_word[0] == "-":
            return
        if curr_word[0] == "/":
            return
        if curr_word[0] == "#":
            return
        if curr_word.startswith("./"):
            return
        if "@" in curr_word and "." in curr_word:
            return
        if curr_word[0] == "@":
            return
        if curr_word.isnumeric():
            return
        if curr_word.isdecimal():
            return
        if curr_word.isspace():
            return

        try:
            if os.path.exists("/bin/" + curr_word):
                return
        except Exception as e:
            pass
        try:
            if os.path.exists("/usr/bin/" + curr_word):
                return
        except Exception as e:
            pass
        try:
            if os.path.exists("/sbin/" + curr_word):
                return
        except Exception as e:
            pass

        if not self.spellChecker.check(curr_word):
            self.env["runtime"]["OutputManager"].present_text(
                _("misspelled"),
                sound_icon="mispell",
                interrupt=False,
                flush=False,
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import datetime
import fcntl
import os
import tempfile
import time

from fenrirscreenreader.core.i18n import _


ANNOUNCEMENT_LOCK_TIMEOUT_SEC = 5.0


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment
        self.last_time = datetime.datetime.now()
        self.lastDateString = ""
        self.lastTimeString = ""

    def shutdown(self):
        pass

    def get_description(self):
        return "No Description found"

    def _get_announcement_lock_path(self):
        return os.path.join(
            tempfile.gettempdir(),
            "fenrirscreenreader-time-announcement.lock",
        )

    def _read_announcement_lock_slot(self, lock_file):
        lock_file.seek(0)
        lock_content = lock_file.readline().strip().split()
        if len(lock_content) < 2:
            return ""
        return lock_content[1]

    def _claim_announcement_lock(self, announcement_slot):
        now = time.time()
        lock_path = self._get_announcement_lock_path()
        try:
            lock_fd = os.open(lock_path, os.O_CREAT | os.O_RDWR, 0o666)
        except OSError:
            return False
        try:
            os.chmod(lock_path, 0o666)
        except OSError:
            pass

        with os.fdopen(lock_fd, "r+", encoding="utf-8") as lock_file:
            try:
                fcntl.flock(lock_file.fileno(), fcntl.LOCK_EX)
                lock_slot = self._read_announcement_lock_slot(lock_file)
                lock_stat = os.fstat(lock_file.fileno())

                if lock_slot == announcement_slot:
                    return False

                lock_file.seek(0)
                lock_file.truncate()
                lock_file.write(f"{os.getpid()} {announcement_slot} {now}\n")
                lock_file.flush()
                os.fsync(lock_file.fileno())
                return True
            finally:
                fcntl.flock(lock_file.fileno(), fcntl.LOCK_UN)

    def run(self):
        if not self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "time", "enabled"
        ):
            return

        on_minutes = self.env["runtime"]["SettingsManager"].get_setting(
            "time", "on_minutes"
        )
        delay_sec = self.env["runtime"]["SettingsManager"].get_setting_as_int(
            "time", "delay_sec"
        )

        # no need
        if on_minutes == "" and delay_sec <= 0:
            return

        on_minutes = on_minutes.split(",")
        now = datetime.datetime.now()

        # ignore on_minutes if there is a delay_sec
        if delay_sec > 0:
            if int((now - self.last_time).total_seconds()) < delay_sec:
                return
            announcement_slot = f"delay:{int(now.timestamp()) // delay_sec}"
        else:
            # should announce?
            if not str(now.minute).zfill(2) in on_minutes:
                return
            # already announced?
            if now.hour == self.last_time.hour:
                if now.minute == self.last_time.minute:
                    return
            announcement_slot = f"minute:{datetime.datetime.strftime(now, '%Y%m%d%H%M')}"

        date_format = self.env["runtime"]["SettingsManager"].get_setting(
            "general", "date_format"
        )
        date_string = datetime.datetime.strftime(now, date_format)

        present_date = (
            self.env["runtime"]["SettingsManager"].get_setting_as_bool(
                "time", "present_date"
            )
            and self.lastDateString != date_string
        )
        present_time = self.env["runtime"][
            "SettingsManager"
        ].get_setting_as_bool("time", "present_time")

        # no changed value to announce
        if not (present_date or present_time):
            return

        if not self._claim_announcement_lock(announcement_slot):
            self.last_time = now
            return

        time_format = self.env["runtime"]["SettingsManager"].get_setting(
            "general", "time_format"
        )
        time_string = datetime.datetime.strftime(now, time_format)

        if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "time", "interrupt"
        ):
            self.env["runtime"]["OutputManager"].interrupt_output()
        if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "time", "announce"
        ):
            self.env["runtime"]["OutputManager"].play_sound_icon("announce")

        if present_time:
            # present the time
            self.env["runtime"]["OutputManager"].present_text(
                _("It's {0}").format(
                    time_string.replace(":00", " O'clock ").lstrip("0")
                ),
                sound_icon="",
                interrupt=False,
            )
            # Check if it's 12:00 AM
            if now.hour == 0 and now.minute == 0:
                # present the date
                self.env["runtime"]["OutputManager"].present_text(
                    date_string, sound_icon="", interrupt=False
                )

        self.last_time = datetime.datetime.now()
        self.lastTimeString = time_string

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import difflib
import time


class TabCompletionManager:
    def __init__(self, timeout=0.5):
        self.timeout = timeout

    def initialize(self, environment):
        self.env = environment
        self._ensure_state()

    def _ensure_state(self):
        if "tabCompletion" not in self.env["commandBuffer"]:
            self.env["commandBuffer"]["tabCompletion"] = {}
        state = self.env["commandBuffer"]["tabCompletion"]
        state.setdefault("pending", None)
        state.setdefault("lastProcessedDelta", "")
        state.setdefault("lastProcessedTime", 0)

    def capture_if_tab(self):
        if not self._last_event_is_tab_press():
            return

        self._ensure_state()
        self.env["commandBuffer"]["tabCompletion"]["pending"] = {
            "timestamp": time.time(),
            "cursor": self.env["screen"]["new_cursor"].copy(),
            "content": self.env["screen"]["new_content_text"],
            "screen": self.env["screen"]["newTTY"],
        }

    def process_update(self):
        self._ensure_state()
        state = self.env["commandBuffer"]["tabCompletion"]
        if self._was_recently_processed(state):
            return ""

        pending = state.get("pending")
        if not pending:
            pending = self._build_pending_from_recent_tab_update()
            if not pending:
                return ""

        if time.time() - pending["timestamp"] > self.timeout:
            state["pending"] = None
            return ""

        if pending.get("screen") != self.env["screen"]["newTTY"]:
            state["pending"] = None
            return ""

        spoken_text = self._get_completion_text(pending)
        if not spoken_text:
            return ""

        state["pending"] = None
        state["lastProcessedDelta"] = self.env["screen"]["new_delta"]
        state["lastProcessedTime"] = time.time()
        return spoken_text

    def _was_recently_processed(self, state):
        if state.get("lastProcessedDelta") != self.env["screen"]["new_delta"]:
            return False
        last_time = state.get("lastProcessedTime")
        if not last_time:
            return False
        return time.time() - last_time <= self.timeout

    def _build_pending_from_recent_tab_update(self):
        if not self._recent_tab_input():
            return None
        return {
            "timestamp": time.time(),
            "cursor": self.env["screen"]["old_cursor"].copy(),
            "content": self.env["screen"]["old_content_text"],
            "screen": self.env["screen"]["newTTY"],
        }

    def _recent_tab_input(self):
        input_manager = self.env["runtime"]["InputManager"]
        if input_manager.get_last_deepest_input() in [["KEY_TAB"]]:
            return True

        last_event = input_manager.get_last_event()
        if not last_event or last_event.get("event_name") != "KEY_TAB":
            return False

        try:
            return (
                time.time() - input_manager.get_last_input_time()
                <= self.timeout
            )
        except Exception:
            return False

    def _last_event_is_tab_press(self):
        input_manager = self.env["runtime"]["InputManager"]
        last_event = input_manager.get_last_event()
        if not last_event:
            return False
        return (
            last_event.get("event_name") == "KEY_TAB"
            and last_event.get("event_state") == 1
        )

    def _get_completion_text(self, pending):
        old_lines = pending["content"].split("\n")
        new_lines = self.env["screen"]["new_content_text"].split("\n")
        cursor = pending["cursor"]
        cursor_y = cursor["y"]

        if cursor_y < len(old_lines) and cursor_y < len(new_lines):
            inserted_text = self._get_cursor_line_inserted_text(
                old_lines[cursor_y],
                new_lines[cursor_y],
                cursor["x"],
                self.env["screen"]["new_cursor"]["x"],
            )
            if inserted_text:
                return self._clean_text(inserted_text)

        candidate_text = self._get_candidate_text(
            old_lines, new_lines, cursor_y
        )
        if candidate_text:
            return self._clean_text(candidate_text)

        delta_text = self.env["screen"]["new_delta"]
        if (
            delta_text
            and not self.env["screen"].get("new_delta_is_typing", False)
        ):
            return self._clean_text(delta_text)

        return ""

    def _get_cursor_line_inserted_text(
        self, old_line, new_line, old_cursor_x, new_cursor_x
    ):
        if new_cursor_x <= old_cursor_x:
            return ""
        if old_line == new_line:
            return ""

        matcher = difflib.SequenceMatcher(
            None, old_line, new_line, autojunk=False
        )
        inserted_parts = []
        for (
            tag,
            old_start,
            old_end,
            new_start,
            new_end,
        ) in matcher.get_opcodes():
            if tag == "equal":
                continue
            if tag == "insert" and old_cursor_x <= new_start <= new_cursor_x:
                inserted_parts.append(new_line[new_start:new_end])
                continue
            if (
                tag == "replace"
                and old_line[old_start:old_end].strip() == ""
                and old_cursor_x <= new_start <= new_cursor_x
            ):
                inserted_parts.append(new_line[new_start:new_end])
                continue
            if (
                tag == "delete"
                and old_line[old_start:old_end].strip() == ""
                and old_start >= old_cursor_x
            ):
                continue
            if old_start >= old_cursor_x:
                return ""

        return "".join(inserted_parts)

    def _get_candidate_text(self, old_lines, new_lines, cursor_y):
        if len(old_lines) != len(new_lines):
            return self._get_inserted_lines(old_lines, new_lines, cursor_y)

        changed_lines = []
        old_cursor_line = (
            old_lines[cursor_y].strip() if cursor_y < len(old_lines) else ""
        )
        for index, old_line in enumerate(old_lines):
            if index == cursor_y:
                continue
            if index < len(new_lines) and old_line != new_lines[index]:
                if new_lines[index].strip() == old_cursor_line:
                    continue
                changed_lines.append(new_lines[index])

        return "\n".join(
            line.rstrip() for line in changed_lines if line.strip()
        )

    def _get_inserted_lines(self, old_lines, new_lines, cursor_y):
        matcher = difflib.SequenceMatcher(
            None, old_lines, new_lines, autojunk=False
        )
        inserted_lines = []
        for (
            tag,
            old_start,
            old_end,
            new_start,
            new_end,
        ) in matcher.get_opcodes():
            if tag not in ["insert", "replace"]:
                continue
            if new_end <= cursor_y:
                continue
            for line in new_lines[new_start:new_end]:
                if line.strip():
                    inserted_lines.append(line.rstrip())
        return "\n".join(inserted_lines)

    def _clean_text(self, text):
        if not text:
            return ""
        if len(text.strip()) != len(text) and text.strip() != "":
            return text.strip()
        return text



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import string
from collections import OrderedDict

from fenrirscreenreader.core import debug

punctuation_data = {
    "LEVELDICT": {
        "none": "",
        "some": "#-$~+*-/\\@",
        "most": ".,:-$~ +*-/\\@!#%^&*()[]}{<>;",
        "all": string.punctuation + " §",
    },
    "PUNCTDICT": {},
    "CUSTOMDICT": OrderedDict(),
    "EMOTICONDICT": OrderedDict(),
}



#!/usr/bin/env python3

import importlib.util
import os
import shutil

from fenrirscreenreader.core.i18n import _

# Load base configuration class
_base_path = os.path.join(os.path.dirname(__file__), "..", "config_base.py")
_spec = importlib.util.spec_from_file_location("config_base", _base_path)
_module = importlib.util.module_from_spec(_spec)
_spec.loader.exec_module(_module)
config_command = _module.config_command


class command(config_command):
    def __init__(self):
        super().__init__()

    def get_description(self):
        return "Reset configuration to default settings"

    def run(self):
        self.present_text("WARNING: This will reset all settings to defaults")
        self.present_text("Creating backup before reset...")

        # Create backup first
        backupSuccess, backupMessage = self.backup_config(announce=False)
        if not backupSuccess:
            self.present_text(
                "Failed to create backup. Reset cancelled for safety.",
                interrupt=False,
            )
            self.play_sound("Error")
            return

        # Find default configuration file
        default_config_path = self.find_default_config()

        if default_config_path and os.path.isfile(default_config_path):
            try:
                # Copy default configuration over current
                shutil.copy2(default_config_path, self.settings_file)

                # Reload the configuration
                self.reload_config()

                self.present_text(
                    "Configuration reset to defaults successfully",
                    interrupt=False,
                    flush=False,
                )
                self.present_text(
                    "Please restart Fenrir for all changes to take effect",
                    interrupt=False,
                    flush=False,
                )

            except Exception as e:
                self.present_text(
                    f"Failed to reset configuration: {str(e)}",
                    interrupt=False,
                    flush=False,
                )
        else:
            # Manually create basic default configuration
            self.create_basic_defaults()

    def find_default_config(self):
        """Find the default configuration file"""
        possible_paths = [
            "/usr/share/fenrirscreenreader/config/settings/settings.conf",
            "/etc/fenrirscreenreader/settings/settings.conf.default",
            os.path.join(
                os.path.dirname(self.settings_file), "settings.conf.default"
            ),
            # Development path
            os.path.join(
                os.path.dirname(__file__),
                "..",
                "..",
                "..",
                "..",
                "..",
                "..",
                "config",
                "settings",
                "settings.conf",
            ),
        ]

        for path in possible_paths:
            if os.path.isfile(path):
                return path

        return None

    def create_basic_defaults(self):
        """Create basic default configuration manually"""
        try:
            self.config.clear()

            # Basic speech defaults
            self.config["speech"] = {
                "enabled": "True",
                "driver": "speechdDriver",
                "rate": "0.5",
                "pitch": "0.5",
                "volume": "1.0",
                "auto_read_incoming": "True",
            }

            # Basic sound defaults
            self.config["sound"] = {
                "enabled": "True",
                "driver": "genericDriver",
                "theme": "default",
                "volume": "0.7",
            }

            # Basic keyboard defaults
            self.config["keyboard"] = {
                "driver": "evdevDriver",
                "device": "ALL",
                "keyboard_layout": "desktop",
                "char_echo_mode": "1",
                "word_echo": "False",
                "char_delete_echo": "True",
            }

            # Basic screen defaults
            self.config["screen"] = {
                "driver": "vcsaDriver",
                "encoding": "auto",
            }

            # Basic general defaults
            self.config["general"] = {
                "punctuation_level": "some",
                "debug_level": "0",
                "number_of_clipboards": "50",
            }

            # Write the configuration
            with open(self.settings_file, "w") as configfile:
                self.config.write(configfile)

            self.present_text(
                "Basic default configuration created",
                interrupt=False,
                flush=False,
            )
            self.present_text(
                "Please restart Fenrir for changes to take effect",
                interrupt=False,
                flush=False,
            )

        except Exception as e:
            self.present_text(
                f"Failed to create default configuration: {str(e)}",
                interrupt=False,
                flush=False,
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import os

from fenrirscreenreader.core import debug
from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("cycles between available keyboard layouts")

    def get_available_layouts(self):
        """Get list of available keyboard layout files"""
        layouts = []

        # Check standard locations for keyboard layouts
        settings_root = "/etc/fenrirscreenreader/"
        if not os.path.exists(settings_root):
            # Fallback to source directory
            import fenrirscreenreader

            fenrir_path = os.path.dirname(fenrirscreenreader.__file__)
            settings_root = fenrir_path + "/../../config/"

        keyboard_path = settings_root + "keyboard/"

        if os.path.exists(keyboard_path):
            for file in os.listdir(keyboard_path):
                if (
                    file.endswith(".conf")
                    and not file.startswith("__")
                    and not file.lower().startswith("pty")
                ):
                    layout_name = file.replace(".conf", "")
                    if layout_name not in layouts:
                        layouts.append(layout_name)

        # Ensure we have at least basic layouts
        if not layouts:
            layouts = ["desktop", "laptop"]
        else:
            layouts.sort()

        return layouts

    def run(self):
        current_layout = self.env["runtime"]["SettingsManager"].get_setting(
            "keyboard", "keyboard_layout"
        )

        # Extract layout name from full path if needed
        if "/" in current_layout:
            current_layout = os.path.basename(current_layout).replace(
                ".conf", ""
            )

        # Get available layouts
        available_layouts = self.get_available_layouts()

        # Find next layout in cycle
        try:
            current_index = available_layouts.index(current_layout)
            next_index = (current_index + 1) % len(available_layouts)
        except ValueError:
            # If current layout not found, start from beginning
            next_index = 0

        next_layout = available_layouts[next_index]

        # Update setting and reload shortcuts
        self.env["runtime"]["SettingsManager"].set_setting(
            "keyboard", "keyboard_layout", next_layout
        )

        # Reload shortcuts with new layout
        try:
            self.env["runtime"]["InputManager"].reload_shortcuts()
            self.env["runtime"]["OutputManager"].present_text(
                _("Switched to {} keyboard layout").format(next_layout),
                interrupt=True,
            )
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "Error reloading shortcuts: " + str(e), debug.DebugLevel.ERROR
            )
            self.env["runtime"]["OutputManager"].present_text(
                _("Error switching keyboard layout"), interrupt=True
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core import debug
from fenrirscreenreader.core.i18n import _


class ScreenDriver:
    def __init__(self):
        self._is_initialized = False
        self.bgColorNames = {
            0: _("black"),
            1: _("blue"),
            2: _("green"),
            3: _("cyan"),
            4: _("red"),
            5: _("Magenta"),
            6: _("brown/yellow"),
            7: _("white"),
        }
        self.fgColorNames = {
            0: _("Black"),
            1: _("Blue"),
            2: _("Green"),
            3: _("Cyan"),
            4: _("Red"),
            5: _("Magenta"),
            6: _("brown/yellow"),
            7: _("Light gray"),
            8: _("Dark gray"),
            9: _("Light blue"),
            10: ("Light green"),
            11: _("Light cyan"),
            12: _("Light red"),
            13: _("Light magenta"),
            14: _("Light yellow"),
            15: _("White"),
        }

    def initialize(self, environment):
        self.env = environment
        self._is_initialized = True

    def shutdown(self):
        self._is_initialized = False

    def get_curr_screen(self):
        pass

    def inject_text_to_screen(self, text, screen=None):
        pass

    def get_curr_application(self):
        pass

    def get_session_information(self):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import os
import signal
import sys
import time

from fenrirscreenreader.core import debug
from fenrirscreenreader.core import remoteInstanceRegistry
from fenrirscreenreader.core import settingsManager
from fenrirscreenreader.core.eventData import FenrirEventType
from fenrirscreenreader.core.i18n import _


class FenrirManager:
    def __init__(self, cliArgs):
        self.is_initialized = False
        self.environment = None
        self.signal_handlers_set = False

        try:
            self.environment = (
                settingsManager.SettingsManager().init_fenrir_config(
                    cliArgs, self
                )
            )
            if not self.environment:
                raise RuntimeError(
                    "Cannot Initialize. Maybe the configfile is not available or not parseable"
                )

            self.environment["runtime"]["OutputManager"].present_text(
                _("Start Fenrir"), sound_icon="ScreenReaderOn", interrupt=True
            )

            # Set signal handlers after successful initialization
            signal.signal(signal.SIGINT, self.capture_signal)
            signal.signal(signal.SIGTERM, self.capture_signal)
            self.signal_handlers_set = True

            self.is_initialized = True
            self.modifierInput = False
            self.singleKeyCommand = False
            self.command = ""
            self.set_process_name()

        except Exception as e:
            # Clean up any partial initialization
            self.cleanup_on_error()
            raise

    def proceed(self):
        if not self.is_initialized:
            return
        self.environment["runtime"]["EventManager"].start_main_event_loop()
        self.shutdown()

    def handle_input(self, event):
        self.environment["runtime"]["DebugManager"].write_debug_out(
            "DEBUG INPUT fenrirMan:" + str(event), debug.DebugLevel.INFO
        )

        if not event["data"]:
            event["data"] = self.environment["runtime"][
                "InputManager"
            ].get_input_event()

        if event["data"]:
            event["data"]["event_name"] = self.environment["runtime"][
                "InputManager"
            ].convert_event_name(event["data"]["event_name"])
            self.environment["runtime"]["InputManager"].handle_input_event(
                event["data"]
            )
        else:
            return

        if self.environment["runtime"]["InputManager"].no_key_pressed():
            self.environment["runtime"]["InputManager"].clear_last_deep_input()

        if self.environment["runtime"]["ScreenManager"].is_ignored_screen():
            self.environment["runtime"]["InputManager"].write_event_buffer()
        else:
            if self.environment["runtime"]["HelpManager"].is_tutorial_mode():
                self.environment["runtime"][
                    "InputManager"
                ].clear_event_buffer()
                self.environment["runtime"]["InputManager"].key_echo(
                    event["data"]
                )

            if self.environment["runtime"]["VmenuManager"].get_active():
                self.environment["runtime"][
                    "InputManager"
                ].clear_event_buffer()
            if self.environment["runtime"]["DiffReviewManager"].is_active():
                self.environment["runtime"][
                    "InputManager"
                ].clear_event_buffer()

            self.detect_shortcut_command()

            if self.modifierInput:
                self.environment["runtime"][
                    "InputManager"
                ].clear_event_buffer()
            if self.singleKeyCommand:
                if self.environment["runtime"][
                    "InputManager"
                ].no_key_pressed():
                    self.environment["runtime"][
                        "InputManager"
                    ].clear_event_buffer()
            else:
                self.environment["runtime"][
                    "InputManager"
                ].write_event_buffer()

        if self.environment["runtime"]["InputManager"].no_key_pressed():
            self.modifierInput = False
            self.singleKeyCommand = False
            self.environment["runtime"]["InputManager"].write_event_buffer()
            self.environment["runtime"]["InputManager"].handle_device_grab()

        self.update_key_forward()

        self.environment["runtime"]["CommandManager"].execute_default_trigger(
            "onKeyInput"
        )

    def handle_execute_command(self, event):
        if not event["data"] or event["data"] == "":
            return
        current_command = event["data"]

        # Check for active read-all and interrupt it on any other key press
        self._interrupt_read_all_if_active(current_command)

        # special modes
        if self.environment["runtime"]["HelpManager"].is_tutorial_mode():
            if self.environment["runtime"]["CommandManager"].command_exists(
                current_command, "help"
            ):
                self.environment["runtime"]["CommandManager"].execute_command(
                    current_command, "help"
                )
                return
        elif self.environment["runtime"]["VmenuManager"].get_active():
            if self.environment["runtime"]["CommandManager"].command_exists(
                current_command, "vmenu-navigation"
            ):
                self.environment["runtime"]["CommandManager"].execute_command(
                    current_command, "vmenu-navigation"
                )
                return

        # default
        self.environment["runtime"]["CommandManager"].execute_command(
            current_command, "commands"
        )

    def _interrupt_read_all_if_active(self, current_command):
        """Check for active read-all and interrupt it on any other key press."""
        try:
            # Skip interruption for read-all commands themselves (they toggle)
            if current_command in ["read_all_by_line", "read_all_by_page"]:
                return
                
            # Stop read-all if any other command is executed
            if ("ReadAllManager" in self.environment["runtime"] and 
                self.environment["runtime"]["ReadAllManager"] and
                self.environment["runtime"]["ReadAllManager"].is_active()):
                
                self.environment["runtime"]["ReadAllManager"].stop_read_all()
                    
        except Exception as e:
            # Ignore errors in interrupt logic to prevent cascading issues
            self.environment["runtime"]["DebugManager"].write_debug_out(
                f"_interrupt_read_all_if_active: {e}",
                debug.DebugLevel.ERROR
            )


    def handle_remote_incomming(self, event):
        if not event["data"]:
            return
        self.environment["runtime"]["RemoteManager"].handle_remote_incomming(
            event["data"]
        )

    def handle_screen_change(self, event):
        self.environment["runtime"]["ScreenManager"].handle_screen_change(
            event["data"]
        )
        if self.environment["runtime"]["VmenuManager"].get_active():
            return
        self.environment["runtime"]["CommandManager"].execute_default_trigger(
            "onScreenChanged"
        )
        self.environment["runtime"]["ScreenDriver"].get_curr_screen()

    def handle_screen_update(self, event):
        self.environment["runtime"]["ScreenManager"].handle_screen_update(
            event["data"]
        )

        if (
            time.time()
            - self.environment["runtime"]["InputManager"].get_last_input_time()
            >= 0.3
        ):
            self.environment["runtime"]["InputManager"].clear_last_deep_input()

        if (
            self.environment["runtime"][
                "CursorManager"
            ].is_cursor_vertical_move()
            or self.environment["runtime"][
                "CursorManager"
            ].is_cursor_horizontal_move()
        ):
            self.environment["runtime"][
                "CommandManager"
            ].execute_default_trigger("onCursorChange")

        self.environment["runtime"]["CommandManager"].execute_default_trigger(
            "onScreenUpdate"
        )
        self.environment["runtime"]["InputManager"].clear_last_deep_input()

    def handle_plug_input_device(self, event):
        try:
            self.environment["runtime"][
                "InputManager"
            ].set_last_detected_devices(event["data"])
        except Exception as e:
            self.environment["runtime"]["DebugManager"].write_debug_out(
                "handle_plug_input_device: Error setting last detected devices: "
                + str(e),
                debug.DebugLevel.ERROR,
            )
        self.environment["runtime"]["InputManager"].handle_plug_input_device(
            event["data"]
        )
        self.environment["runtime"]["CommandManager"].execute_default_trigger(
            "onPlugInputDevice", force=True
        )
        self.environment["runtime"]["InputManager"].set_last_detected_devices(
            None
        )

    def handle_heart_beat(self, event):
        self.environment["runtime"]["CommandManager"].execute_default_trigger(
            "onHeartBeat", force=True
        )

    def detect_shortcut_command(self):
        if self.environment["input"]["key_forward"] != 0:
            return

        if len(self.environment["input"]["prev_input"]) > len(
            self.environment["input"]["curr_input"]
        ):
            return

        if self.environment["runtime"]["InputManager"].is_key_press():
            self.modifierInput = self.environment["runtime"][
                "InputManager"
            ].curr_key_is_modifier()
        else:
            if not self.environment["runtime"][
                "InputManager"
            ].no_key_pressed():
                if self.singleKeyCommand:
                    self.singleKeyCommand = (
                        len(self.environment["input"]["curr_input"]) == 1
                    )

        if not (
            self.singleKeyCommand
            and self.environment["runtime"]["InputManager"].no_key_pressed()
        ):
            current_shortcut = self.environment["runtime"][
                "InputManager"
            ].get_curr_shortcut()
            self.command = self.environment["runtime"][
                "InputManager"
            ].get_command_for_shortcut(current_shortcut)

        if not self.modifierInput:
            if self.environment["runtime"]["InputManager"].is_key_press():
                if self.command != "":
                    self.singleKeyCommand = True
            elif (
                self.environment["runtime"]["DiffReviewManager"].is_active()
                and self.command != ""
            ):
                # Diff mode uses non-Fenrir modified bindings (Shift/Ctrl).
                # Promote resolved shortcuts to executable commands so
                # combinations like Shift+H and Ctrl+Right are dispatched.
                self.singleKeyCommand = True

        if not (self.singleKeyCommand or self.modifierInput):
            return

        # fire event
        if self.command != "":
            if self.modifierInput:
                self.environment["runtime"]["EventManager"].put_to_event_queue(
                    FenrirEventType.execute_command, self.command
                )
                self.command = ""
            else:
                if self.singleKeyCommand:
                    self.environment["runtime"][
                        "EventManager"
                    ].put_to_event_queue(
                        FenrirEventType.execute_command, self.command
                    )
                    self.command = ""

    def update_key_forward(self):
        key_forward = self.environment["input"]["key_forward"]
        input_manager = self.environment["runtime"]["InputManager"]
        if key_forward == -1 and input_manager.no_key_pressed():
            self.environment["input"]["key_forward"] = 1
        elif key_forward == 1 and input_manager.no_key_pressed():
            self.environment["input"]["key_forward"] = 0

    def set_process_name(self, name="fenrir"):
        """Attempts to set the process name to 'fenrir'."""
        try:
            from setproctitle import setproctitle
        except ImportError:
            pass
        else:
            setproctitle(name)
            return True

        try:
            from ctypes import byref
            from ctypes import cdll
            from ctypes import create_string_buffer

            libc = cdll.LoadLibrary("libc.so.6")
            string_buffer = create_string_buffer(len(name) + 1)
            string_buffer.value = bytes(name, "UTF-8")
            libc.prctl(15, byref(string_buffer), 0, 0, 0)
            return True
        except Exception as e:
            self.environment["runtime"]["DebugManager"].write_debug_out(
                "setProcName: Error setting process name: " + str(e),
                debug.DebugLevel.ERROR,
            )

        return False

    def shutdown_request(self):
        try:
            self.environment["runtime"]["EventManager"].stop_main_event_loop()
        except Exception as e:
            self.environment["runtime"]["DebugManager"].write_debug_out(
                "shutdown_request: Error stopping main event loop: " + str(e),
                debug.DebugLevel.ERROR,
            )

    def capture_signal(self, sigInit, frame):
        self.shutdown_request()

    def shutdown(self):
        self.environment["runtime"]["InputManager"].ungrab_all_devices()
        self.environment["runtime"]["EventManager"].stop_main_event_loop()
        self.environment["runtime"]["OutputManager"].present_text(
            _("Quit Fenrir"), sound_icon="ScreenReaderOff", interrupt=True
        )
        self.environment["runtime"]["EventManager"].clean_event_queue()
        time.sleep(0.6)

        for currentManager in self.environment["general"]["managerList"]:
            if self.environment["runtime"][currentManager]:
                self.environment["runtime"][currentManager].shutdown()
                del self.environment["runtime"][currentManager]

        self.environment = None

    def cleanup_on_error(self):
        """Clean up partially initialized state when initialization fails"""
        try:
            # Reset signal handlers to default if they were set
            if self.signal_handlers_set:
                signal.signal(signal.SIGINT, signal.SIG_DFL)
                signal.signal(signal.SIGTERM, signal.SIG_DFL)
                self.signal_handlers_set = False

            # Clean up any initialized managers
            if self.environment:
                try:
                    # Try to ungrab devices if input manager exists
                    if (
                        "runtime" in self.environment
                        and "InputManager" in self.environment["runtime"]
                    ):
                        if self.environment["runtime"]["InputManager"]:
                            self.environment["runtime"][
                                "InputManager"
                            ].ungrab_all_devices()
                except Exception:
                    pass  # Ignore errors during cleanup

                try:
                    # Try to stop event manager if it exists
                    if (
                        "runtime" in self.environment
                        and "EventManager" in self.environment["runtime"]
                    ):
                        if self.environment["runtime"]["EventManager"]:
                            self.environment["runtime"][
                                "EventManager"
                            ].stop_main_event_loop()
                except Exception:
                    pass  # Ignore errors during cleanup

                try:
                    # Try to clean up all managers
                    if (
                        "general" in self.environment
                        and "managerList" in self.environment["general"]
                    ):
                        for currentManager in self.environment["general"][
                            "managerList"
                        ]:
                            if (
                                "runtime" in self.environment
                                and currentManager
                                in self.environment["runtime"]
                                and self.environment["runtime"][currentManager]
                            ):
                                try:
                                    self.environment["runtime"][
                                        currentManager
                                    ].shutdown()
                                    del self.environment["runtime"][
                                        currentManager
                                    ]
                                except Exception:
                                    pass  # Ignore errors during cleanup
                except Exception:
                    pass  # Ignore errors during cleanup

                # Clean up socket files that might not be removed by the driver
                try:
                    socket_file = None
                    screen_driver = None
                    if (
                        "runtime" in self.environment
                        and "SettingsManager" in self.environment["runtime"]
                    ):
                        try:
                            socket_file = self.environment["runtime"][
                                "SettingsManager"
                            ].get_setting("remote", "socket_file")
                        except Exception:
                            pass  # Use default socket file path
                        try:
                            screen_driver = self.environment["runtime"][
                                "SettingsManager"
                            ].get_setting("screen", "driver")
                        except Exception:
                            pass

                    if not socket_file:
                        if screen_driver == "vcsaDriver":
                            socket_file = "/tmp/fenrirscreenreader-deamon.sock"
                            if os.path.exists(socket_file):
                                os.unlink(socket_file)

                        pid_socket_file = (
                            "/tmp/fenrirscreenreader-"
                            + str(os.getpid())
                            + ".sock"
                        )
                        if os.path.exists(pid_socket_file):
                            os.unlink(pid_socket_file)
                    elif os.path.exists(socket_file):
                        os.unlink(socket_file)
                    remoteInstanceRegistry.remove_instance()
                except Exception:
                    pass  # Ignore errors during socket cleanup

                self.environment = None
        except Exception:
            pass  # Ignore all errors during error cleanup



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.utils import word_utils
from fenrirscreenreader.core import debug


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("current word.")

    def run(self):
        self.env["runtime"][
            "CursorManager"
        ].enter_review_mode_curr_text_cursor()

        (
            self.env["screen"]["newCursorReview"]["x"],
            self.env["screen"]["newCursorReview"]["y"],
            curr_word,
            end_of_screen,
            line_break,
        ) = word_utils.get_current_word(
            self.env["screen"]["newCursorReview"]["x"],
            self.env["screen"]["newCursorReview"]["y"],
            self.env["screen"]["new_content_text"],
        )

        # Check if we're in table mode and provide table-aware output
        is_table_mode = self.env["runtime"]["TableManager"].is_table_mode()
        self.env["runtime"]["DebugManager"].write_debug_out(
            f"review_curr_word: is_table_mode={is_table_mode}",
            debug.DebugLevel.INFO
        )
        if is_table_mode:
            # Get current cell info using internal column tracking
            table_info = (
                self.env["runtime"]["TableManager"].get_current_table_cell_info()
            )
            if table_info:
                # Announce with table context - cell content first, then header
                output_text = (
                    f"{table_info['cell_content']} {table_info['column_header']}"
                )
                self.env["runtime"]["OutputManager"].present_text(
                    output_text, interrupt=True, flush=False
                )
                return  # Exit early for table mode
            else:
                # Fallback to regular word announcement
                if curr_word.isspace():
                    self.env["runtime"]["OutputManager"].present_text(
                        _("blank"), interrupt=True, flush=False
                    )
                else:
                    self.env["runtime"]["OutputManager"].present_text(
                        curr_word, interrupt=True, flush=False
                    )
        else:
            # Regular word announcement
            if curr_word.isspace():
                self.env["runtime"]["OutputManager"].present_text(
                    _("blank"), interrupt=True, flush=False
                )
            else:
                self.env["runtime"]["OutputManager"].present_text(
                    curr_word, interrupt=True, flush=False
                )
        if end_of_screen:
            if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
                "review", "end_of_screen"
            ):
                self.env["runtime"]["OutputManager"].present_text(
                    _("end of screen"),
                    interrupt=True,
                    sound_icon="EndOfScreen",
                )
        if line_break:
            if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
                "review", "line_break"
            ):
                self.env["runtime"]["OutputManager"].present_text(
                    _("line break"), interrupt=False, sound_icon="EndOfLine"
                )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.


from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("get previous quick menu value")

    def run(self):
        if self.env["runtime"]["QuickMenuManager"].prev_value():
            value = self.env["runtime"]["QuickMenuManager"].get_current_value()
            self.env["runtime"]["OutputManager"].present_text(
                value, interrupt=True
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.utils import line_utils
from fenrirscreenreader.utils import word_utils


class command:
    def __init__(self):
        self.lastIdent = -1

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return ""

    def run(self):
        if not self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "focus", "cursor"
        ):
            return
        if self.env["runtime"]["ScreenManager"].is_screen_change():
            self.lastIdent = 0
            return

        # Skip if no cursor movement at all
        if (not self.env["runtime"]["CursorManager"].is_cursor_horizontal_move() and 
            not self.env["runtime"]["CursorManager"].is_cursor_vertical_move()):
            return
        x, y, curr_line = line_utils.get_current_line(
            self.env["screen"]["new_cursor"]["x"],
            self.env["screen"]["new_cursor"]["y"],
            self.env["screen"]["new_content_text"],
        )
        curr_ident = self.env["screen"]["new_cursor"]["x"]

        if curr_line.isspace():
            # Don't beep for lines with only spaces - no meaningful indentation
            return
            
        # Lines with actual content - calculate proper indentation
        lastIdent, lastY, last_line = line_utils.get_current_line(
            self.env["screen"]["new_cursor"]["x"],
            self.env["screen"]["new_cursor"]["y"],
            self.env["screen"]["old_content_text"],
        )
        if curr_line.strip() != last_line.strip():
            return
        if len(curr_line.lstrip()) == len(last_line.lstrip()):
            return

        curr_ident = len(curr_line) - len(curr_line.lstrip())

        if curr_ident <= 0:
            return
            
        # Initialize lastIdent if needed
        if self.lastIdent == -1:
            self.lastIdent = curr_ident
            
        # Only beep/announce if indentation level has changed
        if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "general", "auto_present_indent"
        ) and self.lastIdent != curr_ident:
            if self.env["runtime"]["SettingsManager"].get_setting_as_int(
                "general", "auto_present_indent_mode"
            ) in [0, 1]:
                self.env["runtime"]["OutputManager"].play_frequence(
                    curr_ident * 50, 0.1, interrupt=False
                )
            if self.env["runtime"]["SettingsManager"].get_setting_as_int(
                "general", "autePresentIndentMode"
            ) in [0, 2]:
                self.env["runtime"]["OutputManager"].present_text(
                    _("indented ") + str(curr_ident) + " ",
                    interrupt=False,
                    flush=False,
                )
        
        # Always update lastIdent for next comparison
        self.lastIdent = curr_ident

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.


from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("toggles all output settings")

    def run(self):
        if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "speech", "enabled"
        ) or self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "sound", "enabled"
        ):
            self.env["runtime"]["OutputManager"].present_text(
                _("Fenrir muted"), sound_icon="Accept", interrupt=True
            )
            self.env["runtime"]["SettingsManager"].set_setting(
                "speech", "enabled", "False"
            )
            self.env["runtime"]["SettingsManager"].set_setting(
                "sound", "enabled", "False"
            )
        else:
            self.env["runtime"]["SettingsManager"].set_setting(
                "speech", "enabled", "True"
            )
            self.env["runtime"]["SettingsManager"].set_setting(
                "sound", "enabled", "True"
            )
            self.env["runtime"]["OutputManager"].present_text(
                _("Fenrir unmuted"), sound_icon="Cancel", interrupt=True
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python
# -*- encoding: utf-8


from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment
        self.key_macro = [
            [1, "KEY_LEFTCTRL"],
            [1, "KEY_G"],
            [0.05, "SLEEP"],
            [0, "KEY_G"],
            [0, "KEY_LEFTCTRL"],
        ]

    def shutdown(self):
        pass

    def get_description(self):
        return "Learn about the Nano text editor."

    def run(self):
        self.env["runtime"]["OutputManager"].present_text(
            "Okay, loading the information about Nano.", interrupt=True
        )
        self.env["runtime"]["InputManager"].send_keys(self.key_macro)

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import errno
import json
import os
import tempfile
import time


INSTANCE_TIMEOUT_SEC = 30.0


def get_registry_dir():
    return os.path.join(tempfile.gettempdir(), "fenrirscreenreader-instances")


def get_instance_file(pid=None):
    if pid is None:
        pid = os.getpid()
    return os.path.join(get_registry_dir(), f"{pid}.json")


def write_instance(instance_data):
    registry_dir = get_registry_dir()
    os.makedirs(registry_dir, mode=0o755, exist_ok=True)
    os.chmod(registry_dir, 0o755)
    instance_data["updated_at"] = time.time()
    instance_path = get_instance_file(instance_data.get("pid"))
    with open(instance_path, "w", encoding="utf-8") as instance_file:
        json.dump(instance_data, instance_file, sort_keys=True)
        instance_file.write("\n")
    os.chmod(instance_path, 0o644)


def remove_instance(pid=None):
    try:
        os.unlink(get_instance_file(pid))
    except FileNotFoundError:
        pass


def process_exists(pid):
    try:
        os.kill(pid, 0)
        return True
    except OSError as e:
        if e.errno == errno.EPERM:
            return True
        return False


def list_instances():
    registry_dir = get_registry_dir()
    instances = []
    try:
        instance_files = os.listdir(registry_dir)
    except FileNotFoundError:
        return instances

    now = time.time()
    for instance_name in instance_files:
        instance_path = os.path.join(registry_dir, instance_name)
        try:
            with open(instance_path, "r", encoding="utf-8") as instance_file:
                instance_data = json.load(instance_file)
            pid = int(instance_data.get("pid", 0))
            updated_at = float(instance_data.get("updated_at", 0))
        except (OSError, ValueError, TypeError, json.JSONDecodeError):
            continue

        if not pid or not process_exists(pid) or now - updated_at > INSTANCE_TIMEOUT_SEC:
            try:
                os.unlink(instance_path)
            except OSError:
                pass
            continue
        instances.append(instance_data)

    return sorted(instances, key=lambda instance: int(instance.get("pid", 0)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.utils import line_utils


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("current line")

    def run(self):
        self.env["runtime"][
            "CursorManager"
        ].enter_review_mode_curr_text_cursor()

        (
            self.env["screen"]["newCursorReview"]["x"],
            self.env["screen"]["newCursorReview"]["y"],
            curr_line,
        ) = line_utils.get_current_line(
            self.env["screen"]["newCursorReview"]["x"],
            self.env["screen"]["newCursorReview"]["y"],
            self.env["screen"]["new_content_text"],
        )

        if curr_line.isspace():
            self.env["runtime"]["OutputManager"].present_text(
                _("blank"), sound_icon="EmptyLine", interrupt=True, flush=False
            )
        else:
            self.env["runtime"]["OutputManager"].present_text(
                curr_line, interrupt=True, flush=False
            )

    def set_callback(self, callback):
        pass



class command():
    def initialize(self, environment):
        self.env = environment
        self.emoji = "🧛"
    
    def shutdown(self):
        pass
    
    def setCallback(self, callback):
        pass
    
    def getDescription(self):
        return "Vampire emoji"
    
    def run(self):
        self.env["runtime"]["MemoryManager"].add_value_to_first_index(
            "clipboardHistory", self.emoji
        )
        self.env["runtime"]["OutputManager"].present_text(
            "Added vampire to clipboard", 
            interrupt=False, flush=False
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.utils import char_utils


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("presents the current character.")

    def run(self):
        self.env["runtime"][
            "CursorManager"
        ].enter_review_mode_curr_text_cursor()

        # In table mode, sync to cell start if cursor is outside current cell
        if self.env["runtime"]["TableManager"].is_table_mode():
            table_info = self.env["runtime"]["TableManager"].get_current_table_cell_info()
            if table_info:
                cursor_pos = self.env["screen"]["newCursorReview"]
                line_text = self.env["runtime"]["ScreenManager"].get_line_text(
                    cursor_pos["y"]
                )
                if line_text:
                    column_start = self.env["runtime"]["TableManager"].get_column_start_position(
                        line_text, table_info["column_index"]
                    )
                    cell_content = table_info["cell_content"]
                    cell_end = column_start + len(cell_content)
                    
                    # If cursor is outside the current cell, move to cell start
                    if (
                        cursor_pos["x"] < column_start or 
                        cursor_pos["x"] >= cell_end
                    ):
                        self.env["screen"]["newCursorReview"]["x"] = column_start

        (
            self.env["screen"]["newCursorReview"]["x"],
            self.env["screen"]["newCursorReview"]["y"],
            curr_char,
        ) = char_utils.get_current_char(
            self.env["screen"]["newCursorReview"]["x"],
            self.env["screen"]["newCursorReview"]["y"],
            self.env["screen"]["new_content_text"],
        )

        char_utils.present_char_for_review(
            self.env,
            curr_char,
            interrupt=True,
            announce_capital=True,
            flush=False,
        )
        # is has attribute it enabled?
        if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "general", "hasattributes"
        ):
            cursor_pos = self.env["screen"]["newCursorReview"]

            if not self.env["runtime"]["AttributeManager"].has_attributes(
                cursor_pos
            ):
                return
            self.env["runtime"]["OutputManager"].present_text(
                "has attribute", sound_icon="HasAttributes", interrupt=False
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core import debug


def get_text_between_marks(firstMark, secondMark, in_text):
    if in_text is None:
        return ""
    if not isinstance(in_text, list):
        in_text = in_text.split("\n")
    if len(in_text) < 1:
        return ""
    if firstMark is None:
        return ""
    if secondMark is None:
        return ""
    if (firstMark["y"] + 1) * (firstMark["x"] + 1) <= (secondMark["y"] + 1) * (
        secondMark["x"] + 1
    ):
        start_mark = firstMark.copy()
        end_mark = secondMark.copy()
    else:
        end_mark = firstMark.copy()
        start_mark = secondMark.copy()
    text_part = ""
    if start_mark["y"] == end_mark["y"]:
        text_part += in_text[start_mark["y"]][
            start_mark["x"] : end_mark["x"] + 1
        ]
    else:
        curr_y = start_mark["y"]
        while curr_y <= end_mark["y"]:
            if curr_y < end_mark["y"]:
                if curr_y == start_mark["y"]:
                    text_part += in_text[curr_y][start_mark["x"] :]
                else:
                    text_part += in_text[curr_y]
                if len(in_text[curr_y].strip()) != 0:
                    if len(text_part) - len(text_part.rstrip()) > 0:
                        text_part = text_part[: len(text_part.rstrip())] + "\n"
                else:
                    text_part += "\n"
            else:
                text_part += in_text[curr_y][: end_mark["x"] + 1]
            curr_y += 1
    return text_part


def get_text_before_mark(mark, in_text):
    if in_text is None:
        return ""
    if mark is None:
        return ""
    return get_text_between_marks({"x": 0, "y": 0}, mark, in_text)


def get_text_after_mark(mark, in_text):
    if in_text is None:
        return ""
    if mark is None:
        return ""
    in_text = in_text.split("\n")
    return get_text_between_marks(
        mark, {"x": len(in_text[0]) - 1, "y": len(in_text) - 1}, in_text
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.utils import char_utils
from fenrirscreenreader.utils import word_utils


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return ""

    def run(self):
        if not self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "focus", "cursor"
        ):
            return
        if self.env["runtime"]["ScreenManager"].is_screen_change():
            return
        # detect an change on the screen, we just want to cursor arround, so no
        # change should appear
        if self.env["runtime"]["ScreenManager"].is_delta():
            return
        if self.env["runtime"]["ScreenManager"].is_negative_delta():
            return
        # is a vertical change?
        if self.env["runtime"]["CursorManager"].is_cursor_vertical_move():
            return
        # is it a horizontal change?
        if not self.env["runtime"][
            "CursorManager"
        ].is_cursor_horizontal_move():
            return

        # echo word insteed of char
        if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "keyboard", "word_echo"
        ):
            if (
                abs(
                    self.env["screen"]["old_cursor"]["x"]
                    - self.env["screen"]["new_cursor"]["x"]
                )
                != 1
            ):
                # get the word
                new_content = self.env["screen"]["new_content_text"].split(
                    "\n"
                )[self.env["screen"]["new_cursor"]["y"]]
                x, y, curr_word, end_of_screen, line_break = (
                    word_utils.get_current_word(
                        self.env["screen"]["new_cursor"]["x"], 0, new_content
                    )
                )
                if self.env["screen"]["new_cursor"]["x"] == x:
                    return
        x, y, curr_char = char_utils.get_current_char(
            self.env["screen"]["new_cursor"]["x"],
            self.env["screen"]["new_cursor"]["y"],
            self.env["screen"]["new_content_text"],
        )
        # Don't interrupt ongoing auto-read announcements
        do_interrupt = True
        if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "speech", "auto_read_incoming"
        ):
            do_interrupt = False

        if curr_char.isspace():
            # Only announce spaces during pure navigation (arrow keys)
            # Check if this is really navigation by looking at input history
            last_input = self.env["runtime"][
                "InputManager"
            ].get_last_deepest_input()
            if last_input and last_input[0] in [
                "KEY_LEFT",
                "KEY_RIGHT",
                "KEY_UP",
                "KEY_DOWN",
            ]:
                char_utils.present_char_for_review(
                    self.env,
                    curr_char,
                    interrupt=do_interrupt,
                    announce_capital=True,
                    flush=False,
                )
        else:
            self.env["runtime"]["OutputManager"].present_text(
                curr_char,
                interrupt=do_interrupt,
                ignore_punctuation=True,
                announce_capital=True,
                flush=False,
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import importlib.util
import os

from fenrirscreenreader.core.i18n import _

_spec = importlib.util.spec_from_file_location(
    "vmenu_search_base",
    os.path.join(os.path.dirname(__file__), "vmenu_search_base.py"),
)
_module = importlib.util.module_from_spec(_spec)
_spec.loader.exec_module(_module)
vmenu_search_command = _module.vmenu_search_command


class command(vmenu_search_command):
    def __init__(self):
        super().__init__("t")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core import debug

general_data = {
    "args": None,
    "tutorialMode": False,
    "curr_user": "",
    "prev_user": "",
    "managerList": [
        "AttributeManager",
        "PunctuationManager",
        "CursorManager",
        "ApplicationManager",
        "CommandManager",
        "ScreenManager",
        "InputManager",
        "OutputManager",
        "HelpManager",
        "MemoryManager",
        "EventManager",
        "ProcessManager",
        "VmenuManager",
        "QuickMenuManager",
        "ReadAllManager",
        "DiffReviewManager",
        "RemoteManager",
        "SettingsManager",
        "SayAllManager",
        "TableManager",
        "TextManager",
        "BarrierManager",
        "DebugManager",
    ],
    "commandFolderList": [
        "commands",
        "onKeyInput",
        "onCursorChange",
        "onScreenUpdate",
        "onScreenChanged",
        "onHeartBeat",
        "onPlugInputDevice",
        "onApplicationChange",
        "onSwitchApplicationProfile",
        "help",
        "vmenu-navigation",
    ],
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.utils import mark_utils


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("copies marked text to the currently selected clipboard")

    def get_text_from_screen(self, start_mark, end_mark):
        screen_content = self.env["screen"]["new_content_text"]
        screen_lines = screen_content.split("\n")

        start_y = min(start_mark["y"], len(screen_lines) - 1)
        end_y = min(end_mark["y"], len(screen_lines) - 1)

        # If marks are on the same line
        if start_y == end_y:
            line = screen_lines[start_y]
            start_x = min(start_mark["x"], len(line))
            end_x = min(end_mark["x"], len(line)) + 1
            return line[start_x:end_x]

        # Handle multi-line selection
        result = []

        # First line (from start mark to end of line)
        first_line = screen_lines[start_y]
        start_x = min(start_mark["x"], len(first_line))
        result.append(first_line[start_x:].rstrip())

        # Middle lines (complete lines)
        for lineNum in range(start_y + 1, end_y):
            result.append(screen_lines[lineNum].rstrip())

        # Last line (from start to end mark)
        if end_y > start_y:
            last_line = screen_lines[end_y]
            end_x = min(end_mark["x"], len(last_line)) + 1
            result.append(last_line[:end_x].rstrip())

        return "\n".join(result)

    def run(self):
        if not self.env["commandBuffer"]["Marks"]["1"]:
            self.env["runtime"]["OutputManager"].present_text(
                _("One or two marks are needed"), interrupt=True
            )
            return
        if not self.env["commandBuffer"]["Marks"]["2"]:
            self.env["runtime"]["CursorManager"].set_mark()

        # use the last first and the last setted mark as range
        start_mark = self.env["commandBuffer"]["Marks"]["1"].copy()
        end_mark = self.env["commandBuffer"]["Marks"]["2"].copy()

        # Replace mark_utils.get_text_between_marks with our new method
        marked = self.get_text_from_screen(start_mark, end_mark)

        self.env["runtime"]["MemoryManager"].add_value_to_first_index(
            "clipboardHistory", marked
        )
        # reset marks
        self.env["runtime"]["CursorManager"].clear_marks()
        self.env["runtime"]["OutputManager"].present_text(
            marked, sound_icon="CopyToClipboard", interrupt=True
        )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3

from enum import Enum


class DebugLevel(Enum):
    DEACTIVE = 0
    ERROR = 1
    WARNING = 2
    INFO = 3

    def __int__(self):
        return self.value

    def __str__(self):
        return self.name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import _thread
import os
from subprocess import PIPE
from subprocess import Popen

from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment, script_path=""):
        self.env = environment
        self.script_path = script_path

    def shutdown(self):
        pass

    def get_description(self):
        return _("script: {0} fullpath: {1}").format(
            os.path.basename(self.script_path), self.script_path
        )

    def run(self):
        if not os.path.exists(self.script_path):
            self.env["runtime"]["OutputManager"].present_text(
                _("Script file not found"), sound_icon="", interrupt=False
            )
            return
        if not os.path.isfile(self.script_path):
            self.env["runtime"]["OutputManager"].present_text(
                _("Script source is not a valid file"),
                sound_icon="",
                interrupt=False,
            )
            return
        if not os.access(self.script_path, os.X_OK):
            self.env["runtime"]["OutputManager"].present_text(
                _("Script file is not executable"),
                sound_icon="",
                interrupt=False,
            )
            return
        _thread.start_new_thread(self._thread_run, ())

    def _thread_run(self):
        try:
            p = Popen(
                [self.script_path, self.env["general"]["curr_user"]],
                stdout=PIPE,
                stderr=PIPE,
            )
            stdout, stderr = p.communicate()
            stdout = stdout.decode("utf-8")
            stderr = stderr.decode("utf-8")
            self.env["runtime"]["OutputManager"].interrupt_output()
            if stderr != "":
                self.env["runtime"]["OutputManager"].present_text(
                    str(stderr), sound_icon="", interrupt=False
                )
            if stdout != "":
                self.env["runtime"]["OutputManager"].present_text(
                    str(stdout), sound_icon="", interrupt=False
                )
        except Exception as e:
            self.env["runtime"]["OutputManager"].present_text(
                str(e), sound_icon="", interrupt=False
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core import debug
from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return "Detects progress patterns for progress bar monitoring"

    def run(self):
        # Only run if progress monitoring is enabled
        try:
            if (
                "progress_monitoring" in self.env["commandBuffer"]
                and self.env["commandBuffer"]["progress_monitoring"]
            ):
                # Check if current line is a prompt - if so, reset progress
                # state
                if self.is_current_line_prompt():
                    self.reset_progress_state()
                # Only check new incoming text (new_delta), but filter out
                # screen changes
                elif (
                    self.env["screen"]["new_delta"]
                    and not self.env["screen"].get("new_delta_is_typing", False)
                    and self.is_real_progress_update()
                ):
                    self.detect_progress(self.env["screen"]["new_delta"])
        except Exception as e:
            # Log errors for debugging instead of silently ignoring
            self.env["runtime"]["DebugManager"].write_debug_out(
                "Progress detector error: " + str(e),
                debug.DebugLevel.ERROR,
            )

    def is_real_progress_update(self):
        """Check if this is a real progress update vs screen change/window switch"""
        # If the screen/application changed, it's not a progress update
        if self.env["runtime"]["ScreenManager"].is_screen_change():
            self.env["runtime"]["DebugManager"].write_debug_out(
                "Progress filter: screen change detected", debug.DebugLevel.INFO
            )
            return False

        # If there was a large cursor movement, it's likely navigation, not
        # progress
        if self.env["runtime"]["CursorManager"].is_cursor_vertical_move():
            x_move = abs(
                self.env["screen"]["new_cursor"]["x"]
                - self.env["screen"]["old_cursor"]["x"]
            )
            y_move = abs(
                self.env["screen"]["new_cursor"]["y"]
                - self.env["screen"]["old_cursor"]["y"]
            )
            # Large movements suggest navigation, not progress output
            if y_move > 2 or x_move > 20:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    f"Progress filter: large cursor move y={y_move} x={x_move}",
                    debug.DebugLevel.INFO,
                )
                return False

        # Check if delta is too large (screen change) vs small incremental
        # updates
        delta_text = self.env["screen"]["new_delta"]
        delta_length = len(delta_text)
        if (
            delta_length > 200
        ):  # Allow longer progress lines like Claude Code's status
            self.env["runtime"]["DebugManager"].write_debug_out(
                f"Progress filter: delta too long ({delta_length})",
                debug.DebugLevel.INFO,
            )
            return False

        # If delta contains newlines and is substantial, let incoming handler
        # deal with it to avoid interfering with multi-line text output
        if '\n' in delta_text and delta_length > 50:
            self.env["runtime"]["DebugManager"].write_debug_out(
                f"Progress filter: multiline delta ({delta_length} chars)",
                debug.DebugLevel.INFO,
            )
            return False

        # Check if current line looks like a prompt - progress unlikely during
        # prompts
        if self.is_current_line_prompt():
            self.env["runtime"]["DebugManager"].write_debug_out(
                "Progress filter: prompt detected", debug.DebugLevel.INFO
            )
            return False

        return True

    def reset_progress_state(self):
        """Reset progress state when a prompt is detected, allowing new progress operations to start fresh"""
        self.env["runtime"]["DebugManager"].write_debug_out(
            "Resetting progress state due to prompt detection",
            debug.DebugLevel.INFO,
        )
        self.env["commandBuffer"]["lastProgressValue"] = -1
        self.env["commandBuffer"]["lastProgressTime"] = 0

    def detect_progress(self, text):
        import re
        import time

        current_time = time.time()

        # Debug: Print what we're checking
        self.env["runtime"]["DebugManager"].write_debug_out(
            "Progress detector checking: '" + text + "'", debug.DebugLevel.INFO
        )

        # Filter out URLs to prevent false positives
        if self.contains_url(text):
            self.env["runtime"]["DebugManager"].write_debug_out(
                "Skipping progress detection - text contains URL", debug.DebugLevel.INFO
            )
            return

        # Note: Auto-disable on 100% completion removed to respect user
        # settings

        # Pattern 1: Percentage (50%, 25.5%, etc.)
        # Filter out common non-progress percentages (weather, system stats,
        # etc.)
        percent_match = re.search(r"(\d+(?:\.\d+)?)\s*%", text)
        if percent_match:
            percentage = float(percent_match.group(1))
            # Only trigger on realistic progress percentages (0-100%)
            if 0 <= percentage <= 100:
                # Filter out weather/system stats that contain percentages
                if not re.search(
                    r"\b(?:humidity|cpu|memory|disk|usage|temp|weather|forecast)\b",
                    text,
                    re.IGNORECASE,
                ):
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "found percentage: " + str(percentage),
                        debug.DebugLevel.INFO,
                    )
                    if (
                        percentage
                        != self.env["commandBuffer"]["lastProgressValue"]
                    ):
                        self.env["runtime"]["DebugManager"].write_debug_out(
                            "Playing tone for: " + str(percentage),
                            debug.DebugLevel.INFO,
                        )
                        self.play_progress_tone(percentage)
                        self.env["commandBuffer"][
                            "lastProgressValue"
                        ] = percentage
                        self.env["commandBuffer"][
                            "lastProgressTime"
                        ] = current_time
                    return

        # Pattern 1a2: Curl classic progress format (percentage without % symbol)
        # Extract percentage from curl's classic format
        curl_classic_match = re.search(
            r"^\s*(\d+)\s+\d+[kMGT]?\s+(\d+)\s+\d+[kMGT]?\s+\d+\s+\d+\s+\d+[kMGT]?\s+\d+\s+\d+:\d+:\d+\s+\d+:\d+:\d+\s+\d+:\d+:\d+\s+\d+[kMGT]?\s*$", text
        )
        if curl_classic_match:
            # Use the first percentage (total progress)
            percentage = float(curl_classic_match.group(1))
            if 0 <= percentage <= 100:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "found curl classic percentage: " + str(percentage),
                    debug.DebugLevel.INFO,
                )
                if (
                    percentage
                    != self.env["commandBuffer"]["lastProgressValue"]
                ):
                    self.env["runtime"]["DebugManager"].write_debug_out(
                        "Playing tone for curl: " + str(percentage),
                        debug.DebugLevel.INFO,
                    )
                    self.play_progress_tone(percentage)
                    self.env["commandBuffer"][
                        "lastProgressValue"
                    ] = percentage
                    self.env["commandBuffer"][
                        "lastProgressTime"
                    ] = current_time
                return

        # Pattern 1b: Time/token activity (not percentage-based, so use single
        # beep)
        time_match = re.search(r"(?:(?:remaining|elapsed|left|ETA|eta)[:;\s]*(\d+)s|(\d+)s\s+(?:remaining|elapsed|left))", text, re.IGNORECASE)
        token_match = re.search(r"(?:processing|generating|used|consumed)\s+(\d+)\s+tokens", text, re.IGNORECASE)
        # Pattern 1c: dd command output (bytes copied with transfer rate)
        dd_match = re.search(r"\d+\s+bytes.*copied.*\d+\s+s.*[kMGT]?B/s", text)
        # Pattern 1d: Curl-style transfer data (bytes, speed indicators - legacy)
        curl_match = re.search(
            r"(\d+\s+\d+\s+\d+\s+\d+.*?(?:k|M|G)?.*?--:--:--|Speed)", text
        )
        # Pattern 1e: General transfer progress (size, rate, time patterns)
        transfer_match = re.search(
            r"\d+\s+\d+[kMGT]?\s+\d+\s+\d+[kMGT]?.*?\d+\.\d+[kMGT].*?\d+:\d+:\d+", text
        )
        # Pattern 1f: Pacman-style transfer progress (flexible size/speed/time)
        pacman_match = re.search(
            r"\d+(?:\.\d+)?\s+[kKmMgGtT]iB\s+\d+(?:\.\d+)?\s+[kKmMgGtT]iB/s\s+\d+:\d+", text
        )

        if time_match or token_match or dd_match or curl_match or transfer_match or pacman_match:
            # For non-percentage progress, use a single activity beep every 2
            # seconds
            if (
                current_time - self.env["commandBuffer"]["lastProgressTime"]
                >= 2.0
            ):
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "Playing activity beep for transfer progress",
                    debug.DebugLevel.INFO,
                )
                self.play_activity_beep()
                self.env["commandBuffer"]["lastProgressTime"] = current_time
            return

        # Pattern 2: Fraction (15/100, 3 of 10, etc.)
        fraction_match = re.search(r"(\d+)\s*(?:of|/)\s*(\d+)", text)
        if fraction_match:
            current = int(fraction_match.group(1))
            total = int(fraction_match.group(2))
            # Filter out dates, page numbers, and other non-progress fractions
            if (total > 0 and total <= 1000 and current <= total and 
                not re.search(r"\b(?:page|chapter|section|line|row|column|year|month|day)\b", text, re.IGNORECASE) and
                not re.search(r"\d{1,2}/\d{1,2}/\d{2,4}", text)):  # Date pattern
                percentage = (current / total) * 100
                if (
                    percentage
                    != self.env["commandBuffer"]["lastProgressValue"]
                ):
                    self.play_progress_tone(percentage)
                    self.env["commandBuffer"]["lastProgressValue"] = percentage
                    self.env["commandBuffer"][
                        "lastProgressTime"
                    ] = current_time
                return

        # Pattern 3: Progress bars ([####    ], [====>   ], etc.)
        # Improved pattern to avoid matching IRC channels like [#channel]
        bar_match = re.search(r"\[([#=\*]+)([\s\.\-]*)\]", text)
        if bar_match:
            filled = len(bar_match.group(1))
            unfilled = len(bar_match.group(2))
            total = filled + unfilled
            # Require at least 2 progress chars total and unfilled portion must
            # be spaces/dots
            if total >= 2 and (
                not bar_match.group(2)
                or re.match(r"^[\s\.]*$", bar_match.group(2))
            ):
                percentage = (filled / total) * 100
                if (
                    percentage
                    != self.env["commandBuffer"]["lastProgressValue"]
                ):
                    self.play_progress_tone(percentage)
                    self.env["commandBuffer"]["lastProgressValue"] = percentage
                    self.env["commandBuffer"][
                        "lastProgressTime"
                    ] = current_time
                return

        # Pattern 4: Generic activity indicators (Loading..., Working..., etc.)
        activity_pattern = re.search(
            r"(loading|processing|working|installing|downloading|compiling|building).*\.{2,}",
            text,
            re.IGNORECASE,
        )
        if activity_pattern:
            # Play a steady beep every 2 seconds for ongoing activity
            if (
                current_time - self.env["commandBuffer"]["lastProgressTime"]
                >= 2.0
            ):
                self.play_activity_beep()
                self.env["commandBuffer"]["lastProgressTime"] = current_time
            return

        # Pattern 5: Braille spinner indicators
        braille_match = re.search(r'[⠋⠙⠹⠸⠼⠴⠦⠧⠇⠏⡿⣟⣯⣷⣾⣽⣻⢿]', text)
        if braille_match:
            if current_time - self.env["commandBuffer"]["lastProgressTime"] >= 1.0:
                self.play_activity_beep()
                self.env["commandBuffer"]["lastProgressTime"] = current_time
            return

        # Pattern 6: Claude Code working indicators (various symbols + activity text + "esc/ctrl+c to interrupt")
        # Matches any: [symbol] [Task description]… (... to interrupt ...)
        # Symbols include: * ✢ ✽ ✶ ✻ · • ◦ ○ ● ◆ and similar decorative characters
        # Example: ✽ Reviewing script for issues… (ctrl+c to interrupt · 33s · ↑ 1.6k tokens · thought for 4s)
        claude_progress_match = re.search(
            r'[*✢✽✶✻·•◦○●◆]\s+\w+.*?…\s*\(.*(?:esc|ctrl\+c) to interrupt.*\)',
            text,
            re.IGNORECASE,
        )
        if claude_progress_match:
            if current_time - self.env["commandBuffer"]["lastProgressTime"] >= 1.0:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "Playing Claude Code activity beep",
                    debug.DebugLevel.INFO,
                )
                self.play_activity_beep()
                self.env["commandBuffer"]["lastProgressTime"] = current_time
            return

        # Pattern 6b: Claude Code tool invocation indicators (● Tool Name(...))
        # Example: ● Web Search("query here")
        tool_invocation_match = re.search(
            r'[●○◉•◦]\s+(?:Web\s*Search|Read|Write|Edit|Bash|Glob|Grep|Task|WebFetch)\s*\(',
            text,
            re.IGNORECASE,
        )
        if tool_invocation_match:
            if current_time - self.env["commandBuffer"]["lastProgressTime"] >= 1.0:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "Playing Claude Code tool invocation beep",
                    debug.DebugLevel.INFO,
                )
                self.play_activity_beep()
                self.env["commandBuffer"]["lastProgressTime"] = current_time
            return

        # Pattern 6c: Bullet/white bullet activity lines (•/◦ ...)
        bullet_activity_match = re.search(
            (
                r'^\s*[•◦]\s+.*(?:…|\.{3,}|\b(?:thinking|working|processing|'
                r'analyzing|searching|reading|writing|planning|running|'
                r'executing|updating|building|installing|compiling|downloading|'
                r'reviewing|generating|responding|applying|fixing|editing|'
                r'creating|preparing|checking|opening|loading|fetching|'
                r'retrieving|scanning|indexing|summarizing)\b)'
            ),
            text,
            re.IGNORECASE,
        )
        if bullet_activity_match:
            if current_time - self.env["commandBuffer"]["lastProgressTime"] >= 1.0:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "Playing bullet activity beep",
                    debug.DebugLevel.INFO,
                )
                self.play_activity_beep()
                self.env["commandBuffer"]["lastProgressTime"] = current_time
            return

        # Pattern 7: Moon phase spinner indicators
        moon_match = re.search(r'[🌑🌒🌓🌔🌕🌖🌗🌘]', text)
        if moon_match:
            if current_time - self.env["commandBuffer"]["lastProgressTime"] >= 1.0:
                self.play_activity_beep()
                self.env["commandBuffer"]["lastProgressTime"] = current_time
            return

        # Pattern 8: Thinking/processing with timing (🔄 Thinking... 23s)
        thinking_match = re.search(r'🔄[^\w]*(?:thinking|processing|working|analyzing)[^\d]*(\d+)s?\b', text, re.IGNORECASE)
        if thinking_match:
            # Extract timing value for activity beep frequency adjustment
            seconds = int(thinking_match.group(1))
            # Use slightly longer interval for thinking patterns to avoid spam
            thinking_interval = 1.5 if seconds < 10 else 2.0
            if (
                current_time - self.env["commandBuffer"]["lastProgressTime"]
                >= thinking_interval
            ):
                self.env["runtime"]["DebugManager"].write_debug_out(
                    f"Playing thinking activity beep (timing: {seconds}s)",
                    debug.DebugLevel.INFO,
                )
                self.play_activity_beep()
                self.env["commandBuffer"]["lastProgressTime"] = current_time
            return

        # Pattern 9: Half-circle/circle progress indicators (◐ ◓ ◒ ◑)
        circle_match = re.search(r'[◐◓◒◑]', text)
        if circle_match:
            if current_time - self.env["commandBuffer"]["lastProgressTime"] >= 1.0:
                self.play_activity_beep()
                self.env["commandBuffer"]["lastProgressTime"] = current_time
            return

        # Pattern 10: Task status indicators (● Task Output, ○ Task Running, etc.)
        # Matches bullet points with task-related status text
        task_status_match = re.search(
            r'[●○◉]\s+(?:Task\s+)?(?:Output|Running|Pending|Working|Processing)\s+[a-zA-Z0-9]+',
            text,
            re.IGNORECASE,
        )
        if task_status_match:
            if current_time - self.env["commandBuffer"]["lastProgressTime"] >= 1.5:
                self.env["runtime"]["DebugManager"].write_debug_out(
                    "Playing task status activity beep",
                    debug.DebugLevel.INFO,
                )
                self.play_activity_beep()
                self.env["commandBuffer"]["lastProgressTime"] = current_time
            return

    def play_progress_tone(self, percentage):
        # Map 0-100% to 400-1200Hz frequency range
        frequency = 400 + (percentage * 8)
        frequency = max(400, min(1200, frequency))  # Clamp to safe range

        # Use Sox directly for clean quiet tones like: play -qn synth .1 tri
        # 400 gain -8
        self.play_quiet_tone(frequency, 0.1)

    def play_activity_beep(self):
        # Single tone for generic activity
        self.play_quiet_tone(800, 0.08)

    def play_quiet_tone(self, frequency, duration):
        """Play a quiet tone using Sox directly with flood protection"""
        import shlex
        import subprocess
        import time

        # Flood protection: prevent beeps closer than 0.1 seconds apart
        current_time = time.time()
        if not hasattr(self, '_last_beep_time'):
            self._last_beep_time = 0
        
        if current_time - self._last_beep_time < 0.1:
            # Skip this beep to prevent audio crackling on low-resource systems
            return
        
        self._last_beep_time = current_time

        # Build the Sox command: play -qn synth <duration> tri <frequency> gain
        # -8
        command = f"play -qn synth {duration} tri {frequency} gain -8"

        try:
            # Only play if sound is enabled
            if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
                "sound", "enabled"
            ):
                subprocess.Popen(
                    shlex.split(command),
                    stdin=None,
                    stdout=None,
                    stderr=None,
                    shell=False,
                )
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "Sox tone error: " + str(e), debug.DebugLevel.ERROR
            )

    def is_current_line_prompt(self):
        """Check if the current line looks like a standalone prompt (not command with progress)"""
        import re

        try:
            # Get the current screen content
            if not self.env["screen"]["new_content_text"]:
                return False

            lines = self.env["screen"]["new_content_text"].split("\n")
            if not lines:
                return False

            # Check the last line (most common) and current cursor line for
            # prompt patterns
            lines_to_check = []

            # Add last line (most common for prompts)
            if lines:
                lines_to_check.append(lines[-1])

            # Add current cursor line if different from last line
            if (
                self.env["screen"]["new_cursor"]["y"] < len(lines)
                and self.env["screen"]["new_cursor"]["y"] != len(lines) - 1
            ):
                lines_to_check.append(
                    lines[self.env["screen"]["new_cursor"]["y"]]
                )

            # Standalone prompt patterns (no commands mixed in)
            standalone_prompt_patterns = [
                r"^\s*\$\s*$",  # Just $ (with whitespace)
                r"^\s*#\s*$",  # Just # (with whitespace)
                r"^\s*>\s*$",  # Just > (with whitespace)
                r"^\[.*\]\s*[\\\$#>]\s*$",  # [path]$ without commands
                r"^[a-zA-Z0-9._-]+[\\\$#>]\s*$",  # bash-5.1$ without commands
                # Interactive prompt patterns (these ARE standalone)
                r".*\?\s*\[[YyNn]/[YyNn]\]\s*$",  # ? [Y/n] or ? [y/N] style
                r".*\?\s*\[[Yy]es/[Nn]o\]\s*$",  # ? [Yes/No] style
                # "continue? [Y/n]" style
                r".*continue\?\s*\[[YyNn]/[YyNn]\].*$",
                r"^::.*\?\s*\[[YyNn]/[YyNn]\].*$",  # pacman style prompts
                # Authentication prompts (these ARE standalone)
                r"^\[[Ss]udo\]\s*[Pp]assword\s*for\s+.*:\s*$",  # [sudo] password
                r"^[Pp]assword\s*:\s*$",  # Password:
                r".*[Pp]assword\s*:\s*$",  # general password prompts
                # Continuation prompts (these ARE standalone)
                r"^[Pp]ress\s+any\s+key\s+to\s+continue.*$",  # Press any key
                r"^[Aa]re\s+you\s+sure\?\s*.*$",  # Are you sure?
            ]

            for line in lines_to_check:
                line = line.strip()
                if not line:
                    continue

                # Check if this line contains both a prompt AND other content (like commands)
                # If so, don't treat it as a standalone prompt
                has_prompt_marker = bool(
                    re.search(r".*@.*[\\\$#>]", line)
                    or re.search(r"^\[.*\]\s*[\\\$#>]", line)
                )
                if has_prompt_marker:
                    # If line has prompt marker but also has significant content after it,
                    # it's a command line, not a standalone prompt
                    prompt_end = max(
                        line.rfind("$"),
                        line.rfind("#"),
                        line.rfind(">"),
                        line.rfind("\\"),
                    )
                    if (
                        prompt_end >= 0 and prompt_end < len(line) - 5
                    ):  # More than just whitespace after prompt
                        continue  # This is a command line, not a standalone prompt

                for pattern in standalone_prompt_patterns:
                    try:
                        if re.search(pattern, line):
                            return True
                    except re.error:
                        continue

            return False

        except Exception:
            # If anything fails, assume it's not a prompt to be safe
            return False

    def contains_url(self, text):
        """Check if text contains URLs that might cause false progress detection"""
        import re
        
        # Specific URL patterns - only match actual URLs, not filenames
        url_patterns = [
            r"\S+://\S+\.\S{2,}",         # Any protocol:// with domain.ext
            r"www\.[^\s]+\.[a-zA-Z]{2,}",  # www.domain.ext patterns
        ]
        
        for pattern in url_patterns:
            if re.search(pattern, text, re.IGNORECASE):
                return True
        return False

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.


from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        self.env["runtime"]["HelpManager"].toggle_tutorial_mode()
        return _(
            "Exiting tutorial mode. To enter tutorial mode again press Fenrir+f1"
        )

    def run(self):
        self.env["runtime"]["HelpManager"].toggle_tutorial_mode()
        if self.env["runtime"]["HelpManager"].is_tutorial_mode():
            self.env["runtime"]["OutputManager"].present_text(
                _(
                    "Entering tutorial mode. In this mode commands are described but not "
                    "executed. You can move through the list of commands with the up and "
                    "down arrow keys. To Exit tutorial mode press Fenrir+f1."
                ),
                interrupt=True,
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.utils import mark_utils


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _(
            "Presents the currently selected text that will be copied to the "
            "clipboard"
        )

    def run(self):
        if not (
            self.env["commandBuffer"]["Marks"]["1"]
            and self.env["commandBuffer"]["Marks"]["2"]
        ):
            self.env["runtime"]["OutputManager"].present_text(
                _("please set begin and endmark"), interrupt=True
            )
            return

        # use the last first and the last setted mark as range
        start_mark = self.env["commandBuffer"]["Marks"]["1"].copy()
        end_mark = self.env["commandBuffer"]["Marks"]["2"].copy()

        marked = mark_utils.get_text_between_marks(
            start_mark, end_mark, self.env["screen"]["new_content_text"]
        )

        if marked.isspace():
            self.env["runtime"]["OutputManager"].present_text(
                _("blank"), sound_icon="EmptyLine", interrupt=True
            )
        else:
            self.env["runtime"]["OutputManager"].present_text(
                marked, interrupt=True
            )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.


from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return "No description found"

    def run(self):
        if self.env["input"]["old_num_lock"] == self.env["input"]["new_num_lock"]:
            return

        # Only announce numlock changes if an actual numlock key was pressed
        # AND the LED state actually changed (some numpads send spurious NUMLOCK events)
        current_input = self.env["input"]["curr_input"]
        previous_input = self.env["input"]["prev_input"]
        relevant_input = current_input or previous_input

        # Check if this is a genuine numlock key press by verifying:
        # 1. KEY_NUMLOCK is in the current input sequence
        # 2. The LED state has actually changed
        # 3. This isn't just a side effect from a KP_ key (which some buggy numpads do)
        is_genuine_numlock = (
            relevant_input and
            "KEY_NUMLOCK" in relevant_input and
            not any(
                key.startswith("KEY_KP")
                for key in relevant_input
                if isinstance(key, str)
            )
        )

        if is_genuine_numlock:
            if self.env["input"]["new_num_lock"]:
                self.env["runtime"]["OutputManager"].present_text(
                    _("Numlock on"), interrupt=True
                )
            else:
                self.env["runtime"]["OutputManager"].present_text(
                    _("Numlock off"), interrupt=True
                )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3

from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return "Voice browser: exit browser mode"

    def run(self):
        if "voiceBrowserInstance" in self.env["runtime"]:
            self.env["runtime"]["voiceBrowserInstance"].exit_voice_browser()

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import signal
import time
from multiprocessing import Process
from threading import Thread

from fenrirscreenreader.core import debug
from fenrirscreenreader.core.eventData import FenrirEventType


# Standalone functions for multiprocessing (cannot be instance methods)
def _heart_beat_timer(running):
    """
    Standalone heartbeat timer function for multiprocessing.
    Returns current timestamp after a short sleep.
    """
    try:
        time.sleep(0.5)
    except Exception as e:
        print(f"ProcessManager _heart_beat_timer: Error during sleep: {e}")
    return time.time()


def _custom_event_worker_process(
    running, event_queue, function, pargs=None, run_once=False
):
    """
    Standalone worker function for custom events in multiprocessing.
    Cannot use instance methods due to pickle limitations with self.env.
    """
    if not callable(function):
        return
    while running.value:
        try:
            if pargs:
                function(running, event_queue, pargs)
            else:
                function(running, event_queue)
        except Exception as e:
            # Cannot use DebugManager in multiprocess context
            print(
                f"ProcessManager:_custom_event_worker_process:function("
                f"{function}):{e}"
            )
        if run_once:
            break


def _simple_event_worker_process(
    running, event_queue, event, function, pargs=None, run_once=False
):
    """
    Standalone worker function for simple events in multiprocessing.
    Cannot use instance methods due to pickle limitations with self.env.
    """
    if not isinstance(event, FenrirEventType):
        return
    if not callable(function):
        return
    while running.value:
        data = None
        try:
            if pargs:
                data = function(running, pargs)
            else:
                data = function(running)
        except Exception as e:
            # Cannot use DebugManager in multiprocess context
            print(
                f"ProcessManager:_simple_event_worker_process:function("
                f"{function}):{e}"
            )
        try:
            event_queue.put({"Type": event, "data": data}, timeout=0.1)
        except Exception as e:
            print(f"ProcessManager: Failed to put event to queue: {e}")
        if run_once:
            break


class ProcessManager:
    def __init__(self):
        self._Processes = []
        self._Threads = []

    def initialize(self, environment):
        self.env = environment
        self.running = self.env["runtime"]["EventManager"].get_running()
        self.add_simple_event_thread(
            FenrirEventType.heart_beat,
            _heart_beat_timer,
            multiprocess=True,
        )

    def shutdown(self):
        self.terminate_all_processes()

    def terminate_all_processes(self):
        for proc in self._Processes:
            # try:
            #    proc.terminate()
            # except KeyboardInterrupt:
            #    pass
            # except:
            #    pass
            proc.join(timeout=5.0)  # Timeout to prevent hanging shutdown
        for t in self._Threads:
            t.join(timeout=5.0)  # Timeout to prevent hanging shutdown

    def heart_beat_timer(self, active):
        try:
            time.sleep(0.5)
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "ProcessManager heart_beat_timer: Error during sleep: "
                + str(e),
                debug.DebugLevel.ERROR,
            )
        return time.time()

    def add_custom_event_thread(
        self, function, pargs=None, multiprocess=False, run_once=False
    ):
        event_queue = self.env["runtime"]["EventManager"].get_event_queue()
        original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
        if multiprocess:
            t = Process(
                target=_custom_event_worker_process,
                args=(self.running, event_queue, function, pargs, run_once),
            )
            self._Processes.append(t)
        else:  # use thread instead of process
            t = Thread(
                target=self.custom_event_worker_thread,
                args=(event_queue, function, pargs, run_once),
            )
            self._Threads.append(t)
        t.start()
        signal.signal(signal.SIGINT, original_sigint_handler)

    def add_simple_event_thread(
        self, event, function, pargs=None, multiprocess=False, run_once=False
    ):
        original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
        if multiprocess:
            # Get event queue reference before creating process
            event_queue = self.env["runtime"]["EventManager"].get_event_queue()
            t = Process(
                target=_simple_event_worker_process,
                args=(self.running, event_queue, event, function, pargs, run_once),
            )
            self._Processes.append(t)
        else:
            t = Thread(
                target=self.simple_event_worker_thread,
                args=(event, function, pargs, run_once),
            )
            self._Threads.append(t)
        t.start()
        signal.signal(signal.SIGINT, original_sigint_handler)

    def custom_event_worker_thread(
        self, event_queue, function, pargs=None, run_once=False
    ):
        # if not isinstance(event_queue, Queue):
        #    return
        if not callable(function):
            return
        while self.running.value:
            try:
                if pargs:
                    function(self.running, event_queue, pargs)
                else:
                    function(self.running, event_queue)
            except Exception as e:
                # Cannot use DebugManager in multiprocess context due to
                # pickle limitations with file handles
                print(
                    "ProcessManager:custom_event_worker_thread:function("
                    + str(function)
                    + "):"
                    + str(e)
                )
            if run_once:
                break

    def simple_event_worker_thread(
        self, event, function, pargs=None, run_once=False
    ):
        if not isinstance(event, FenrirEventType):
            return
        if not callable(function):
            return
        while self.running.value:
            data = None
            try:
                if pargs:
                    data = function(self.running, pargs)
                else:
                    data = function(self.running)
            except Exception as e:
                # Cannot use DebugManager in multiprocess context due to
                # pickle limitations with file handles
                print(
                    "ProcessManager:simple_event_worker_thread:function("
                    + str(function)
                    + "):"
                    + str(e)
                )
            self.env["runtime"]["EventManager"].put_to_event_queue(event, data)
            if run_once:
                break



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import time

from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return _("Announces large text insertions at the cursor")

    def run(self):
        if self.env["runtime"]["ScreenManager"].is_screen_change():
            return
        if not self.env["runtime"]["ScreenManager"].is_delta():
            return

        x_move = (
            self.env["screen"]["new_cursor"]["x"]
            - self.env["screen"]["old_cursor"]["x"]
        )
        if x_move < 5:
            return
        if self._is_recent_tab_input():
            return

        delta_text = self.env["screen"]["new_delta"]
        if not self._matches_cursor_insert(x_move, delta_text):
            return

        curr_delta = delta_text
        if (
            len(curr_delta.strip()) != len(curr_delta)
            and curr_delta.strip() != ""
        ):
            curr_delta = curr_delta.strip()
        if not curr_delta:
            return

        do_interrupt = True
        if self.env["runtime"]["SettingsManager"].get_setting_as_bool(
            "speech", "auto_read_incoming"
        ):
            do_interrupt = False

        self.env["runtime"]["OutputManager"].present_text(
            curr_delta,
            interrupt=do_interrupt,
            announce_capital=True,
            flush=False,
        )

    def _is_recent_tab_input(self):
        input_manager = self.env["runtime"].get("InputManager")
        if not input_manager:
            return False
        if input_manager.get_last_deepest_input() in [["KEY_TAB"]]:
            return True

        last_event = input_manager.get_last_event()
        if not last_event or last_event.get("event_name") != "KEY_TAB":
            return False
        try:
            return time.time() - input_manager.get_last_input_time() <= 0.5
        except Exception:
            return False

    def _matches_cursor_insert(self, x_move, delta_text):
        if not delta_text or "\n" in delta_text:
            return False

        delta_len = len(delta_text)
        if x_move == delta_len:
            return True
        if abs(x_move - delta_len) <= 2:
            return True
        return delta_len > 10 and abs(x_move - delta_len) <= (delta_len * 0.2)

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import string

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.utils import word_utils

initialized = False
try:
    import enchant

    initialized = True
except Exception as e:
    pass


class command:
    def __init__(self):
        self.language = ""
        self.spellChecker = None

    def initialize(self, environment):
        self.env = environment
        self.update_spell_language()

    def shutdown(self):
        pass

    def get_description(self):
        return _("removes the current word from the exceptions dictionary")

    def update_spell_language(self):
        self.spellChecker = enchant.Dict(
            self.env["runtime"]["SettingsManager"].get_setting(
                "general", "spell_check_language"
            )
        )
        self.language = self.env["runtime"]["SettingsManager"].get_setting(
            "general", "spell_check_language"
        )

    def run(self):
        if not initialized:
            self.env["runtime"]["OutputManager"].present_text(
                _("pyenchant is not installed"), interrupt=True
            )
            return
        if (
            self.env["runtime"]["SettingsManager"].get_setting(
                "general", "spell_check_language"
            )
            != self.language
        ):
            try:
                self.update_spell_language()
            except Exception as e:
                return

        cursor_pos = self.env["runtime"][
            "CursorManager"
        ].get_review_or_text_cursor()

        # get the word
        new_content = self.env["screen"]["new_content_text"].split("\n")[
            cursor_pos["y"]
        ]
        x, y, curr_word, end_of_screen, line_break = (
            word_utils.get_current_word(cursor_pos["x"], 0, new_content)
        )
        curr_word = curr_word.strip(
            string.whitespace + r'!"#$%&()*+,-./:;<=Â?@[\\]^_{|}~'
        )
        if not curr_word.isspace():
            if self.spellChecker.is_removed(curr_word):
                self.env["runtime"]["OutputManager"].present_text(
                    _("{0} is not in the dictionary").format(
                        curr_word,
                    ),
                    sound_icon="Cancel",
                    interrupt=True,
                )
            else:
                self.spellChecker.remove(curr_word)
                self.env["runtime"]["OutputManager"].present_text(
                    _("{0} removed").format(
                        curr_word,
                    ),
                    sound_icon="Accept",
                    interrupt=True,
                )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.utils import line_utils
from fenrirscreenreader.utils import mark_utils


class bookmark_command:
    """Base class for bookmark operations - read, set, clear"""

    def __init__(self, bookmark_id, action="read"):
        self.ID = str(bookmark_id)
        self.action = action

    def initialize(self, environment):
        self.env = environment
        # Always initialize bookmark structure - all commands need this
        if self.ID not in self.env["commandBuffer"]["bookMarks"]:
            self.env["commandBuffer"]["bookMarks"][self.ID] = {}

    def shutdown(self):
        pass

    def get_description(self):
        if self.action == "read":
            return _("read Bookmark {0}").format(self.ID)
        elif self.action == "set":
            return _("set Bookmark {0}").format(self.ID)
        elif self.action == "clear":
            return _("remove Bookmark {0}").format(self.ID)
        return f"{self.action} Bookmark {self.ID}"

    def run(self):
        if self.action == "read":
            self._read_bookmark()
        elif self.action == "set":
            self._set_bookmark()
        elif self.action == "clear":
            self._clear_bookmark()

    def _read_bookmark(self):
        curr_app = self.env["runtime"][
            "ApplicationManager"
        ].get_current_application()

        if not self.env["commandBuffer"]["bookMarks"][self.ID]:
            self.env["runtime"]["OutputManager"].present_text(
                "Bookmark {0} not set".format(self.ID), interrupt=True
            )
            return

        if curr_app not in self.env["commandBuffer"]["bookMarks"][self.ID]:
            self.env["runtime"]["OutputManager"].present_text(
                "Bookmark for application {0} not set".format(curr_app),
                interrupt=True,
            )
            return

        if not self.env["commandBuffer"]["bookMarks"][self.ID][curr_app].get(
            "1"
        ):
            self.env["runtime"]["OutputManager"].present_text(
                "Bookmark for application {0} not set".format(curr_app),
                interrupt=True,
            )
            return

        # Get bookmarked text
        marked = ""
        start_mark = self.env["commandBuffer"]["bookMarks"][self.ID][curr_app][
            "1"
        ].copy()

        if self.env["commandBuffer"]["bookMarks"][self.ID][curr_app]["2"]:
            end_mark = self.env["commandBuffer"]["bookMarks"][self.ID][
                curr_app
            ]["2"].copy()
            marked = mark_utils.get_text_between_marks(
                start_mark, end_mark, self.env["screen"]["new_content_text"]
            )
        else:
            x, y, marked = line_utils.get_current_line(
                start_mark["x"],
                start_mark["y"],
                self.env["screen"]["new_content_text"],
            )

        if marked.isspace():
            self.env["runtime"]["OutputManager"].present_text(
                _("blank"), sound_icon="EmptyLine", interrupt=True
            )
        else:
            self.env["runtime"]["OutputManager"].present_text(
                marked, interrupt=True
            )

    def _set_bookmark(self):
        if not self.env["commandBuffer"]["Marks"]["1"]:
            self.env["runtime"]["OutputManager"].present_text(
                _("No mark found"), interrupt=True
            )
            return

        curr_app = self.env["runtime"][
            "ApplicationManager"
        ].get_current_application()
        self.env["commandBuffer"]["bookMarks"][self.ID][curr_app] = {}

        bookmarks = self.env["commandBuffer"]["bookMarks"][self.ID][curr_app]
        bookmarks["1"] = self.env["commandBuffer"]["Marks"]["1"].copy()

        if self.env["commandBuffer"]["Marks"]["2"]:
            bookmarks["2"] = self.env["commandBuffer"]["Marks"]["2"].copy()
        else:
            bookmarks["2"] = None

        self.env["runtime"]["OutputManager"].present_text(
            _("Bookmark {0} set for application {1}").format(
                self.ID, curr_app
            ),
            interrupt=True,
        )

        # Clear marks after setting bookmark
        self.env["commandBuffer"]["Marks"]["1"] = None
        self.env["commandBuffer"]["Marks"]["2"] = None

    def _clear_bookmark(self):
        curr_app = self.env["runtime"][
            "ApplicationManager"
        ].get_current_application()

        bookmarks = self.env["commandBuffer"]["bookMarks"]
        if self.ID in bookmarks and curr_app in bookmarks[self.ID]:
            del self.env["commandBuffer"]["bookMarks"][self.ID][curr_app]
            self.env["runtime"]["OutputManager"].present_text(
                _("Bookmark {0} removed for application {1}").format(
                    self.ID, curr_app
                ),
                interrupt=True,
            )
        else:
            self.env["runtime"]["OutputManager"].present_text(
                _("Bookmark {0} not set for application {1}").format(
                    self.ID, curr_app
                ),
                interrupt=True,
            )

    def set_callback(self, callback):
        pass


# Factory function to create bookmark command instances


def create_bookmark_commands():
    """Create all bookmark command instances"""
    commands = {}

    # Create read bookmark commands (bookmark_1 through bookmark_10)
    for i in range(1, 11):
        commands[f"bookmark_{i}"] = lambda i=i: bookmark_command(i, "read")

    # Create set bookmark commands (set_bookmark_1 through set_bookmark_10)
    for i in range(1, 11):
        commands[f"set_bookmark_{i}"] = lambda i=i: bookmark_command(i, "set")

    # Create clear bookmark commands (clear_bookmark_1 through
    # clear_bookmark_10)
    for i in range(1, 11):
        commands[f"clear_bookmark_{i}"] = lambda i=i: bookmark_command(
            i, "clear"
        )

    return commands


# For backwards compatibility, provide individual command classes
# This allows the existing command loading system to work unchanged


def _make_command_class(bookmark_id, action):
    """Create a command class for a specific bookmark and action"""

    class command(bookmark_command):
        def __init__(self):
            super().__init__(bookmark_id, action)

    return command


# Generate individual command classes for each bookmark operation
# These will be used by the existing command loading system


# Read bookmarks (bookmark_1.py style)
for i in range(1, 11):
    globals()[f"bookmark_{i}_command"] = _make_command_class(i, "read")

# Set bookmarks (set_bookmark_1.py style)
for i in range(1, 11):
    globals()[f"set_bookmark_{i}_command"] = _make_command_class(i, "set")

# Clear bookmarks (clear_bookmark_1.py style)
for i in range(1, 11):
    globals()[f"clear_bookmark_{i}_command"] = _make_command_class(i, "clear")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import time

from fenrirscreenreader.core import debug

screen_data = {
    "columns": 0,
    "lines": 0,
    "oldDelta": "",
    "oldAttribDelta": "",
    "oldNegativeDelta": "",
    "oldCursorReview": None,
    "oldCursorAttrib": None,
    "old_cursor": {"x": 0, "y": 0},
    "oldContentBytes": b"",
    "old_content_text": "",
    "oldContentAttrib": None,
    "oldApplication": "",
    "oldTTY": None,
    "new_delta": "",
    "new_delta_is_typing": False,
    "newNegativeDelta": "",
    "newAttribDelta": "",
    "newCursorReview": None,
    "newCursorAttrib": None,
    "new_cursor": {"x": 0, "y": 0},
    "newContentBytes": b"",
    "new_content_text": "",
    "newContentAttrib": None,
    "newTTY": "0",
    "new_application": "",
    "lastScreenUpdate": time.time(),
    "autoIgnoreScreens": [],
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.


from fenrirscreenreader.core.i18n import _


class command:
    def __init__(self):
        pass

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def get_description(self):
        return "No Description found"

    def run(self):
        self.env["runtime"]["OutputManager"].present_text(
            _("screen {0}").format(self.env["screen"]["newTTY"]),
            sound_icon="ChangeTTY",
            interrupt=True,
            flush=False,
        )
        self.env["runtime"]["OutputManager"].present_text(
            self.env["screen"]["new_content_text"],
            interrupt=False,
            flush=False,
        )

    def set_callback(self, callback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import time

from fenrirscreenreader.core import debug


class InputDriver:
    def __init__(self):
        self._initialized = False

    def initialize(self, environment):
        self.env = environment
        self._is_initialized = True

    def shutdown(self):
        if self._initialized:
            self.remove_all_devices()
        self._is_initialized = False

    def get_input_event(self):
        time.sleep(0.1)
        return None

    def clear_event_buffer(self):
        if not self._initialized:
            return
        del self.env["input"]["event_buffer"][:]

    def update_input_devices(self, new_devices=None, init=False):
        if not self._initialized:
            return

    def get_led_state(self, led=0):
        if not self._initialized:
            return False
        return False

    def toggle_led_state(self, led=0):
        if not self._initialized:
            return

    def grab_all_devices(self):
        if not self._initialized:
            return True
        return True

    def ungrab_all_devices(self):
        if not self._initialized:
            return True
        return True

    def force_ungrab(self):
        """Emergency method to release grabbed devices in case of failure"""
        if not self._initialized:
            return True
        try:
            # Try standard ungrab first
            return self.ungrab_all_devices()
        except Exception as e:
            # Just log the failure and inform the user
            if (
                hasattr(self, "env")
                and "runtime" in self.env
                and "DebugManager" in self.env["runtime"]
            ):
                self.env["runtime"]["DebugManager"].write_debug_out(
                    f"Emergency device release failed: {str(e)}",
                    debug.DebugLevel.ERROR,
                )
            else:
                print(f"Emergency device release failed: {str(e)}")
        return False

    def has_i_devices(self):
        if not self._initialized:
            return False
        return True

    def remove_all_devices(self):
        if not self._initialized:
            return

    def send_key(self):
        if not self._initialized:
            return

    def __del__(self):
        if not self._initialized:
            return
        try:
            self.remove_all_devices()
        except Exception as e:
            print(f"InputDriver __del__: Error removing devices: {str(e)}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.
# speech-dispatcher driver

from fenrirscreenreader.core import debug
from fenrirscreenreader.core.speechDriver import speech_driver


class driver(speech_driver):
    """Speech-dispatcher driver for Fenrir screen reader.
    
    This driver provides text-to-speech functionality through 
    speech-dispatcher,
    which acts as a common interface to various TTS engines. It supports voice
    selection, speech parameters (rate, pitch, volume), and multiple TTS 
    modules.
    
    Features:
    - Dynamic voice switching and parameter adjustment
    - Support for multiple speech-dispatcher modules (espeak, festival, etc.)
    - Real-time speech cancellation and queueing
    - Language and voice selection
    - Speech callbacks for synchronization
    
    Attributes:
        _sd: Speech-dispatcher client connection
        language (str): Current speech language
        voice (str): Current voice name
        module (str): Current TTS module name
    """
    def __init__(self):
        speech_driver.__init__(self)

    def initialize(self, environment):
        """Initialize the speech-dispatcher connection.
        
        Establishes connection to speech-dispatcher daemon and configures
        initial speech parameters. Sets up callbacks and prepares the
        speech subsystem for use.
        
        Args:
            environment: Fenrir environment dictionary with settings and managers
            
        Note:
            Gracefully handles cases where speech-dispatcher is not available.
        """
        self._sd = None
        self.env = environment
        self._is_initialized = False

        # Only set these if they haven't been set yet (preserve existing
        # values)
        if not hasattr(self, "language") or self.language is None:
            self.language = ""
        if not hasattr(self, "voice") or self.voice is None:
            self.voice = ""
        if not hasattr(self, "module") or self.module is None:
            self.module = ""

        try:
            import speechd

            self._sd = speechd.SSIPClient("fenrir-dev")
            self._punct = speechd.PunctuationMode()
            self._is_initialized = True
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "SpeechDriver initialize:" + str(e), debug.DebugLevel.ERROR
            )

    def shutdown(self):
        """Shutdown the speech-dispatcher connection.
        
        Cleanly closes the connection to speech-dispatcher and releases
        any allocated resources.
        """
        if not self._is_initialized:
            return
        self.cancel()
        try:
            self._sd.close()
        except Exception as e:
            pass
        self._is_initialized = False

    def speak(self, text, queueable=True, ignore_punctuation=False):
        """Speak the given text through speech-dispatcher.
        
        Args:
            text (str): Text to speak
            queueable (bool): Whether speech can be queued with other speech
            ignore_punctuation (bool): Whether to ignore punctuation settings
            
        Note:
            Handles text preprocessing and manages speech queue based on parameters.
        """
        if not queueable:
            self.cancel()
        if not self._is_initialized:
            self.initialize(self.env)
            if not queueable:
                self.cancel()
            if not self._is_initialized:
                return

        try:
            if self.module != "":
                self._sd.set_output_module(self.module)
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "SpeechDriver set_module:" + str(e), debug.DebugLevel.ERROR
            )

        try:
            if self.language != "":
                self._sd.set_language(self.language)
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "SpeechDriver set_language:" + str(e), debug.DebugLevel.ERROR
            )

        try:
            if self.voice != "":
                self._sd.set_synthesis_voice(self.voice)
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "SpeechDriver set_voice:" + str(e), debug.DebugLevel.ERROR
            )

        try:
            if ignore_punctuation:
                self._sd.set_punctuation(self._punct.ALL)
            else:
                self._sd.set_punctuation(self._punct.NONE)
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "SpeechDriver set_punctuation:" + str(e),
                debug.DebugLevel.ERROR,
            )

        try:
            # Check if read-all is active and add callback if needed
            if ("ReadAllManager" in self.env["runtime"] and 
                self.env["runtime"]["ReadAllManager"] and 
                self.env["runtime"]["ReadAllManager"].is_active()):
                
                import speechd
                
                def read_all_callback(msg_id, client_id, event_type, *args):
                    if (event_type == speechd.CallbackType.END and 
                        self.env["runtime"]["ReadAllManager"] and
                        self.env["runtime"]["ReadAllManager"].is_active()):
                        self.env["runtime"]["ReadAllManager"].speech_completed()
                
                self._sd.speak(text, callback=read_all_callback, 
                             event_types=(speechd.CallbackType.END,))
            else:
                self._sd.speak(text)
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "SpeechDriver speak:" + str(e), debug.DebugLevel.ERROR
            )
            self._is_initialized = False

    def cancel(self):
        """Cancel all pending and current speech.
        
        Immediately stops speech output and clears the speech queue.
        """
        if not self._is_initialized:
            self.initialize(self.env)
            if not self._is_initialized:
                return
        try:
            self._sd.cancel()
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "SpeechDriver cancel:" + str(e), debug.DebugLevel.ERROR
            )
            self._is_initialized = False

    def set_pitch(self, pitch):
        """Set the speech pitch.
        
        Args:
            pitch (float): Speech pitch (0.0 to 1.0, where 0.5 is normal)
        """
        if not self._is_initialized:
            return
        try:
            self._sd.set_pitch(int(-100 + pitch * 200))
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "SpeechDriver set_pitch:" + str(e), debug.DebugLevel.ERROR
            )

    def set_rate(self, rate):
        """Set the speech rate.
        
        Args:
            rate (float): Speech rate (0.0 to 1.0, where 0.5 is normal)
        """
        if not self._is_initialized:
            return
        try:
            self._sd.set_rate(int(-100 + rate * 200))
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "SpeechDriver set_rate:" + str(e), debug.DebugLevel.ERROR
            )

    def set_volume(self, volume):
        """Set the speech volume.
        
        Args:
            volume (float): Volume level (0.0 to 1.0)
        """
        if not self._is_initialized:
            return
        try:
            self._sd.set_volume(int(-100 + volume * 200))
        except Exception as e:
            self.env["runtime"]["DebugManager"].write_debug_out(
                "SpeechDriver set_volume:" + str(e), debug.DebugLevel.ERROR
            )




#!/usr/bin/env python3

import importlib.util
import os

from fenrirscreenreader.core.i18n import _

# Load base configuration class
_base_path = os.path.join(os.path.dirname(__file__), "..", "config_base.py")
_spec = importlib.util.spec_from_file_location("config_base", _base_path)
_module = importlib.util.module_from_spec(_spec)
_spec.loader.exec_module(_module)
config_command = _module.config_command


class command(config_command):
    def __init__(self):
        super().__init__()

    def get_description(self):
        return "Toggle word echo when pressing space"

    def run(self):
        current_state = self.getBooleanSetting("keyboard", "word_echo", False)
        new_state = self.toggleBooleanSetting("keyboard", "word_echo")

        if new_state != current_state:
            state_text = "enabled" if new_state else "disabled"
            self.present_text(f"Word echo {state_text}")
            if new_state:
                self.present_text("Words will be spoken when you press space")
            else:
                self.present_text(
                    "Words will not be spoken when you press space"
                )
            self.play_sound("Accept")
        else:
            self.present_text("Failed to change word echo setting")
            self.play_sound("Error")
